#!/usr/bin/env python
"""Provides conversion functions to be used during RDFProtoStruct migration."""

from grr_response_core.lib.rdfvalues import artifacts as rdf_artifacts
from grr_response_proto import artifact_pb2
from grr_response_proto import flows_pb2


def ToProtoArtifactSource(
    rdf: rdf_artifacts.ArtifactSource,
) -> artifact_pb2.ArtifactSource:
  return rdf.AsPrimitiveProto()


def ToRDFArtifactSource(
    proto: artifact_pb2.ArtifactSource,
) -> rdf_artifacts.ArtifactSource:
  return rdf_artifacts.ArtifactSource.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoArtifact(rdf: rdf_artifacts.Artifact) -> artifact_pb2.Artifact:
  return rdf.AsPrimitiveProto()


def ToRDFArtifact(proto: artifact_pb2.Artifact) -> rdf_artifacts.Artifact:
  return rdf_artifacts.Artifact.FromSerializedBytes(proto.SerializeToString())


def ToProtoArtifactDescriptor(
    rdf: rdf_artifacts.ArtifactDescriptor,
) -> artifact_pb2.ArtifactDescriptor:
  return rdf.AsPrimitiveProto()


def ToRDFArtifactDescriptor(
    proto: artifact_pb2.ArtifactDescriptor,
) -> rdf_artifacts.ArtifactDescriptor:
  return rdf_artifacts.ArtifactDescriptor.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoArtifactCollectorFlowArgs(
    rdf: rdf_artifacts.ArtifactCollectorFlowArgs,
) -> flows_pb2.ArtifactCollectorFlowArgs:
  return rdf.AsPrimitiveProto()


def ToRDFArtifactCollectorFlowArgs(
    proto: flows_pb2.ArtifactCollectorFlowArgs,
) -> rdf_artifacts.ArtifactCollectorFlowArgs:
  return rdf_artifacts.ArtifactCollectorFlowArgs.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoArtifactProgress(
    rdf: rdf_artifacts.ArtifactProgress,
) -> flows_pb2.ArtifactProgress:
  return rdf.AsPrimitiveProto()


def ToRDFArtifactProgress(
    proto: flows_pb2.ArtifactProgress,
) -> rdf_artifacts.ArtifactProgress:
  return rdf_artifacts.ArtifactProgress.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoArtifactCollectorFlowProgress(
    rdf: rdf_artifacts.ArtifactCollectorFlowProgress,
) -> flows_pb2.ArtifactCollectorFlowProgress:
  return rdf.AsPrimitiveProto()


def ToRDFArtifactCollectorFlowProgress(
    proto: flows_pb2.ArtifactCollectorFlowProgress,
) -> rdf_artifacts.ArtifactCollectorFlowProgress:
  return rdf_artifacts.ArtifactCollectorFlowProgress.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoClientActionResult(
    rdf: rdf_artifacts.ClientActionResult,
) -> artifact_pb2.ClientActionResult:
  return rdf.AsPrimitiveProto()


def ToRDFClientActionResult(
    proto: artifact_pb2.ClientActionResult,
) -> rdf_artifacts.ClientActionResult:
  return rdf_artifacts.ClientActionResult.FromSerializedBytes(
      proto.SerializeToString()
  )



#!/usr/bin/env python
"""The various Dummy example rdfvalues."""

from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_proto import dummy_pb2


class DummyRequest(rdf_structs.RDFProtoStruct):
  """Request for Dummy action."""

  protobuf = dummy_pb2.DummyRequest
  rdf_deps = []


class DummyResult(rdf_structs.RDFProtoStruct):
  """Result for Dummy action."""

  protobuf = dummy_pb2.DummyResult
  rdf_deps = []



#!/usr/bin/env python
import io
import os
import random
import stat as stat_mode
import time

from absl.testing import absltest

from grr_response_core.lib.rdfvalues import timeline as rdf_timeline
from grr_response_core.lib.util import statx
from grr_response_core.lib.util import temp
from grr_response_proto import timeline_pb2


class TimelineEntryTest(absltest.TestCase):

  def testFromStat(self):
    with temp.AutoTempDirPath(remove_non_empty=True) as dirpath:
      filepath = os.path.join(dirpath, "foobar")

      # We subtract a second to account for the terrible Windows clock accuracy.
      time_before_ns = time.time_ns() - 1e9

      with io.open(filepath, mode="wb") as filedesc:
        filedesc.write(b"1234567")

      # We add a second to account for the terrible Windows clock accuracy.
      time_after_ns = time.time_ns() + 1e9

      # TODO(hanuszczak): `AutoTempFilePath` should return a `Path` object.
      filepath_bytes = filepath.encode("utf-8")
      filepath_stat = os.lstat(filepath)

      entry = rdf_timeline.TimelineEntry.FromStat(filepath_bytes, filepath_stat)

      self.assertEqual(entry.size, 7)
      self.assertTrue(stat_mode.S_ISREG(entry.mode))

      self.assertBetween(entry.atime_ns, time_before_ns, time_after_ns)
      self.assertBetween(entry.mtime_ns, time_before_ns, time_after_ns)
      self.assertBetween(entry.ctime_ns, time_before_ns, time_after_ns)

      self.assertEqual(entry.dev, filepath_stat.st_dev)
      self.assertEqual(entry.ino, filepath_stat.st_ino)
      self.assertEqual(entry.uid, filepath_stat.st_uid)
      self.assertEqual(entry.gid, filepath_stat.st_gid)

  def testFromStatx(self):
    with temp.AutoTempFilePath() as filepath:
      filepath = filepath.encode("utf-8")

      with open(filepath, mode="wb") as filedesc:
        filedesc.write(b"1234567")

      statx_result = statx.Get(filepath)
      entry = rdf_timeline.TimelineEntry.FromStatx(filepath, statx_result)

      self.assertEqual(entry.size, 7)
      self.assertTrue(stat_mode.S_ISREG(entry.mode))

      self.assertEqual(entry.dev, statx_result.dev)
      self.assertEqual(entry.ino, statx_result.ino)
      self.assertEqual(entry.uid, statx_result.uid)
      self.assertEqual(entry.gid, statx_result.gid)
      self.assertEqual(entry.attributes, statx_result.attributes)

      self.assertEqual(entry.atime_ns, statx_result.atime_ns)
      self.assertEqual(entry.btime_ns, statx_result.btime_ns)
      self.assertEqual(entry.ctime_ns, statx_result.ctime_ns)
      self.assertEqual(entry.mtime_ns, statx_result.mtime_ns)

  def testSerializeAndDeserializeStream(self):
    serialize = rdf_timeline.SerializeTimelineEntryStream
    deserialize = rdf_timeline.DeserializeTimelineEntryStream

    def RandomEntry() -> timeline_pb2.TimelineEntry:
      entry = timeline_pb2.TimelineEntry()
      entry.path = os.urandom(4096)
      entry.mode = random.randint(int(0x0000), int(0xFFFF - 1))
      entry.size = random.randint(0, int(1e9))
      return entry

    entries = [RandomEntry() for _ in range(3000)]

    self.assertEqual(list(deserialize(serialize(entries))), entries)


if __name__ == "__main__":
  absltest.main()



#!/usr/bin/env python
"""Provides conversion functions to be used during RDFProtoStruct migration."""

from grr_response_core.lib.rdfvalues import stats as rdf_stats
from grr_response_proto import analysis_pb2
from grr_response_proto import jobs_pb2


def ToProtoDistribution(rdf: rdf_stats.Distribution) -> jobs_pb2.Distribution:
  return rdf.AsPrimitiveProto()


def ToRDFDistribution(proto: jobs_pb2.Distribution) -> rdf_stats.Distribution:
  return rdf_stats.Distribution.FromSerializedBytes(proto.SerializeToString())


def ToProtoMetricFieldDefinition(
    rdf: rdf_stats.MetricFieldDefinition,
) -> jobs_pb2.MetricFieldDefinition:
  return rdf.AsPrimitiveProto()


def ToRDFMetricFieldDefinition(
    proto: jobs_pb2.MetricFieldDefinition,
) -> rdf_stats.MetricFieldDefinition:
  return rdf_stats.MetricFieldDefinition.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoMetricMetadata(
    rdf: rdf_stats.MetricMetadata,
) -> jobs_pb2.MetricMetadata:
  return rdf.AsPrimitiveProto()


def ToRDFMetricMetadata(
    proto: jobs_pb2.MetricMetadata,
) -> rdf_stats.MetricMetadata:
  return rdf_stats.MetricMetadata.FromSerializedBytes(proto.SerializeToString())


def ToProtoStatsHistogramBin(
    rdf: rdf_stats.StatsHistogramBin,
) -> jobs_pb2.StatsHistogramBin:
  return rdf.AsPrimitiveProto()


def ToRDFStatsHistogramBin(
    proto: jobs_pb2.StatsHistogramBin,
) -> rdf_stats.StatsHistogramBin:
  return rdf_stats.StatsHistogramBin.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoStatsHistogram(
    rdf: rdf_stats.StatsHistogram,
) -> jobs_pb2.StatsHistogram:
  return rdf.AsPrimitiveProto()


def ToRDFStatsHistogram(
    proto: jobs_pb2.StatsHistogram,
) -> rdf_stats.StatsHistogram:
  return rdf_stats.StatsHistogram.FromSerializedBytes(proto.SerializeToString())


def ToProtoRunningStats(rdf: rdf_stats.RunningStats) -> jobs_pb2.RunningStats:
  return rdf.AsPrimitiveProto()


def ToRDFRunningStats(proto: jobs_pb2.RunningStats) -> rdf_stats.RunningStats:
  return rdf_stats.RunningStats.FromSerializedBytes(proto.SerializeToString())


def ToProtoClientResourcesStats(
    rdf: rdf_stats.ClientResourcesStats,
) -> jobs_pb2.ClientResourcesStats:
  return rdf.AsPrimitiveProto()


def ToRDFClientResourcesStats(
    proto: jobs_pb2.ClientResourcesStats,
) -> rdf_stats.ClientResourcesStats:
  return rdf_stats.ClientResourcesStats.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoSample(rdf: rdf_stats.Sample) -> analysis_pb2.Sample:
  return rdf.AsPrimitiveProto()


def ToRDFSample(proto: analysis_pb2.Sample) -> rdf_stats.Sample:
  return rdf_stats.Sample.FromSerializedBytes(proto.SerializeToString())


def ToProtoSampleFloat(rdf: rdf_stats.SampleFloat) -> analysis_pb2.SampleFloat:
  return rdf.AsPrimitiveProto()


def ToRDFSampleFloat(proto: analysis_pb2.SampleFloat) -> rdf_stats.SampleFloat:
  return rdf_stats.SampleFloat.FromSerializedBytes(proto.SerializeToString())


def ToProtoGraph(rdf: rdf_stats.Graph) -> analysis_pb2.Graph:
  return rdf.AsPrimitiveProto()


def ToRDFGraph(proto: analysis_pb2.Graph) -> rdf_stats.Graph:
  return rdf_stats.Graph.FromSerializedBytes(proto.SerializeToString())


def ToProtoClientGraphSeries(
    rdf: rdf_stats.ClientGraphSeries,
) -> analysis_pb2.ClientGraphSeries:
  return rdf.AsPrimitiveProto()


def ToRDFClientGraphSeries(
    proto: analysis_pb2.ClientGraphSeries,
) -> rdf_stats.ClientGraphSeries:
  return rdf_stats.ClientGraphSeries.FromSerializedBytes(
      proto.SerializeToString()
  )



#!/usr/bin/env python
from absl.testing import absltest

from grr_response_core.lib.rdfvalues import mig_protodict
from grr_response_proto import jobs_pb2


class FromProtoDictToNativeDictTest(absltest.TestCase):

  def testFromProtoDictToNativeDict(self):
    expected = {"a": 1, "b": 2, "c": "d"}
    proto = jobs_pb2.Dict(
        dat=[
            jobs_pb2.KeyValue(
                k=jobs_pb2.DataBlob(string="a"), v=jobs_pb2.DataBlob(integer=1)
            ),
            jobs_pb2.KeyValue(
                k=jobs_pb2.DataBlob(string="b"), v=jobs_pb2.DataBlob(integer=2)
            ),
            jobs_pb2.KeyValue(
                k=jobs_pb2.DataBlob(string="c"), v=jobs_pb2.DataBlob(string="d")
            ),
        ]
    )
    got = mig_protodict.FromProtoDictToNativeDict(proto)
    self.assertEqual(expected, got)


class FromNativeDictToProtoDictTest(absltest.TestCase):

  def testFromNativeDictToProtoDict(self):
    native = {"a": 1, "b": 2, "c": "d"}
    expected = jobs_pb2.Dict(
        dat=[
            jobs_pb2.KeyValue(
                k=jobs_pb2.DataBlob(string="a"), v=jobs_pb2.DataBlob(integer=1)
            ),
            jobs_pb2.KeyValue(
                k=jobs_pb2.DataBlob(string="b"), v=jobs_pb2.DataBlob(integer=2)
            ),
            jobs_pb2.KeyValue(
                k=jobs_pb2.DataBlob(string="c"), v=jobs_pb2.DataBlob(string="d")
            ),
        ]
    )
    got = mig_protodict.FromNativeDictToProtoDict(native)
    self.assertEqual(expected, got)


class FromProtoAttributedDictToNativeDictTest(absltest.TestCase):

  def testFromProtoAttributedDictToNativeDict(self):
    expected = {"a": 1, "b": 2, "c": "d"}
    proto = jobs_pb2.AttributedDict(
        dat=[
            jobs_pb2.KeyValue(
                k=jobs_pb2.DataBlob(string="a"), v=jobs_pb2.DataBlob(integer=1)
            ),
            jobs_pb2.KeyValue(
                k=jobs_pb2.DataBlob(string="b"), v=jobs_pb2.DataBlob(integer=2)
            ),
            jobs_pb2.KeyValue(
                k=jobs_pb2.DataBlob(string="c"), v=jobs_pb2.DataBlob(string="d")
            ),
        ]
    )
    got = mig_protodict.FromProtoAttributedDictToNativeDict(proto)
    self.assertEqual(expected, got)


class FromNativeDictToProtoAttributedDictTest(absltest.TestCase):

  def testFromNativeDictToProtoAttributedDict(self):
    native = {"a": 1, "b": 2, "c": "d"}
    expected = jobs_pb2.AttributedDict(
        dat=[
            jobs_pb2.KeyValue(
                k=jobs_pb2.DataBlob(string="a"), v=jobs_pb2.DataBlob(integer=1)
            ),
            jobs_pb2.KeyValue(
                k=jobs_pb2.DataBlob(string="b"), v=jobs_pb2.DataBlob(integer=2)
            ),
            jobs_pb2.KeyValue(
                k=jobs_pb2.DataBlob(string="c"), v=jobs_pb2.DataBlob(string="d")
            ),
        ]
    )
    got = mig_protodict.FromNativeDictToProtoAttributedDict(native)
    self.assertEqual(expected, got)


if __name__ == "__main__":
  absltest.main()



#!/usr/bin/env python
"""Tests for the DefaultStatsCollector."""

from absl import app

from grr_response_core.stats import default_stats_collector
from grr_response_core.stats import stats_test_utils
from grr.test_lib import test_lib


class DefaultStatsCollectorTest(stats_test_utils.StatsCollectorTest):

  def _CreateStatsCollector(self):
    return default_stats_collector.DefaultStatsCollector()


def main(argv):
  test_lib.main(argv)


if __name__ == "__main__":
  app.run(main)



#!/usr/bin/env python
"""Crypto rdfvalue tests."""

import binascii
import hashlib
import os
import unittest
from unittest import mock

from absl import app

from grr_response_core import config
from grr_response_core.lib import config_lib
from grr_response_core.lib import rdfvalue
from grr_response_core.lib import type_info
from grr_response_core.lib import utils
from grr_response_core.lib.rdfvalues import crypto as rdf_crypto
from grr_response_core.lib.rdfvalues import test_base as rdf_test_base
from grr_response_proto import jobs_pb2
from grr.test_lib import test_lib


class SignedBlobTest(rdf_test_base.RDFValueTestMixin, test_lib.GRRBaseTest):
  rdfvalue_class = rdf_crypto.SignedBlob

  def setUp(self):
    super().setUp()
    self.private_key = config.CONFIG[
        "PrivateKeys.executable_signing_private_key"
    ]
    self.public_key = config.CONFIG["Client.executable_signing_public_key"]

  def GenerateSample(self, number=0):
    result = self.rdfvalue_class()
    result.Sign(("Sample %s" % number).encode("ascii"), self.private_key)

    return result

  @unittest.skip(
      "Samples are expected to be different for the same data as PSS padding"
      " generates a new Hash for every sample."
  )
  def testComparisons(self):
    pass

  def testSignVerify(self):
    sample = self.GenerateSample()

    self.assertTrue(sample.Verify(self.public_key))

    # Change the data - this should fail since the hash is incorrect.
    sample.data += b"X"
    self.assertRaises(
        rdf_crypto.VerificationError, sample.Verify, self.public_key
    )

    # Update the hash
    sample.digest = hashlib.sha256(sample.data).digest()

    # Should still fail.
    self.assertRaises(
        rdf_crypto.VerificationError, sample.Verify, self.public_key
    )

    # If we change the digest verification should fail.
    sample = self.GenerateSample()
    sample.digest_type = sample.HashType.MD5

    self.assertRaises(rdfvalue.DecodeError, sample.Verify, self.public_key)

    # PSS should be accepted.
    sample = self.GenerateSample()
    sample.signature_type = sample.SignatureType.RSA_PSS
    sample.signature = self.private_key.Sign(sample.data)
    sample.Verify(self.public_key)


class CryptoTestBase(test_lib.GRRBaseTest):
  """Base class for all crypto tests here."""


class TestCryptoTypeInfos(CryptoTestBase):
  """Test that invalid configuration types are rejected.

  There is no need to check for success here because if that did not work we
  would not be able to run any tests.
  """

  def setUp(self):
    super().setUp()
    config_stubber = test_lib.PreserveConfig()
    config_stubber.Start()
    self.addCleanup(config_stubber.Stop)

  def testInvalidRSAPrivateKey(self):
    """Deliberately try to parse invalid RSA keys."""
    config.CONFIG.Initialize(data="""
[PrivateKeys]
executable_signing_private_key = -----BEGIN RSA PRIVATE KEY-----
        MIIBOgIBAAJBALnfFW1FffeKPs5PLUhFOSkNrr9TDCODQAI3WluLh0sW7/ro93eo
        -----END RSA PRIVATE KEY-----
""")
    config.CONFIG.context = []

    with self.assertRaises(config_lib.ConfigFormatError):
      config.CONFIG.Get("PrivateKeys.executable_signing_private_key")

  def testRSAPublicKeySuccess(self):
    config.CONFIG.Initialize(data="""
[Client]
executable_signing_public_key = -----BEGIN PUBLIC KEY-----
    MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAzbuwYnKTTleU2F4zu1gI
    /BolzR74470j6wK7QrQp5b5QkmfevCdX540Ax3mFt6isChrhIXy4LmTvM5SeiCvs
    6su8ro7ZYqtbZjpoBWorESy5VAnnuiuwiAvI7IOPYbSu8e+mHQa893hLYv9q0kQe
    1CtlvQsKir8aQ0YC63UWhrcMcMYCjUaRkWo8JUWoCW/GuZjbuFF0ZZsA13cVSOhF
    g1dxvX4/+FXlnyZQbu2Aex4NU9NGYetl+SP10Uq/c92wZ34ENU3rH8Kw8Tar/dku
    g9+hgrcgt7JRECXmho+WFrKM1H6z1CNe6uEALMfogNl9Mm2jCgavl1wbV92pIKT1
    pQIDAQAB
    -----END PUBLIC KEY-----
""")
    config.CONFIG.context = []

    errors = config.CONFIG.Validate("Client")
    self.assertFalse(errors)

  def testRSAPublicKeyFailure(self):
    """Deliberately try to parse an invalid public key."""
    config.CONFIG.Initialize(data="""
[Client]
executable_signing_public_key = -----BEGIN PUBLIC KEY-----
        GpJgTFkTIAgX0Ih5lxoFB5TUjUfJFbBkSmKQPRA/IyuLBtCLQgwkTNkCAwEAAQ==
        -----END PUBLIC KEY-----
""")
    config.CONFIG.context = []

    errors = config.CONFIG.Validate("Client")
    self.assertCountEqual(
        list(errors.keys()), ["Client.executable_signing_public_key"]
    )

  def testRSAPrivate(self):
    """Tests parsing an RSA private key."""
    config.CONFIG.Initialize(data="""
[PrivateKeys]
executable_signing_private_key = -----BEGIN RSA PRIVATE KEY-----
    MIIEowIBAAKCAQEAzbuwYnKTTleU2F4zu1gI/BolzR74470j6wK7QrQp5b5Qkmfe
    vCdX540Ax3mFt6isChrhIXy4LmTvM5SeiCvs6su8ro7ZYqtbZjpoBWorESy5VAnn
    uiuwiAvI7IOPYbSu8e+mHQa893hLYv9q0kQe1CtlvQsKir8aQ0YC63UWhrcMcMYC
    jUaRkWo8JUWoCW/GuZjbuFF0ZZsA13cVSOhFg1dxvX4/+FXlnyZQbu2Aex4NU9NG
    Yetl+SP10Uq/c92wZ34ENU3rH8Kw8Tar/dkug9+hgrcgt7JRECXmho+WFrKM1H6z
    1CNe6uEALMfogNl9Mm2jCgavl1wbV92pIKT1pQIDAQABAoIBAAz6jaCm+T+HO2rg
    myizClK7JI4u6W/Wj1+Z+xh0wgg4wXT8VrBt1qK912Io0AR7gH+dB/2t6aGgTk3Q
    Tvr+CGgcj8E/CPGWUqxdkWaMDCj9IQ5El148N4VcTEyNanwfuFdSHBuMgLB89zEz
    i03grHA2nP24Jq4E6yIk1nYmW2mGoZSIrrSDG/VQj4NipeGxbwJsOGdFmzYlefk9
    lwC7Db9v2ocdytxlH7lxhVqEnOokLXT7VO/EPTyF3CcObZx2JG2P4d6KILcsRePg
    +xMTp0wmq4S3Hoiwi2+RZNH7aHn5igvcNbx0+Eu9CIKINuNwmw3c/2bAz6u48FUY
    8aUmQkECgYEA7cD3E8TqUTxQxCrcuHKVzbS5ThLvfSbIo7hctz2b9zshomRaeh1m
    qg3p5shQ0+p029H26mu8oi5e130QCmzbnhWoZZbVAY4SEe6+5SAF1d1Xhwn8jTZ5
    Vqo5B7ILKVvxHWkgdYFV9yi6c23/kSzFDZXW4QFt3eZgPv8wLxWU3uUCgYEA3YWh
    VcX2UDrQR8r4Qy9YHuQ85vW6O5QKvSz//Ckw9Z6TfFOf5DRmAHyt2IpnmmbeuhDD
    SsnXv1fsqJWcbl+/nP2cNuIhLXnUwebtTW3Qo9PSi7APhOYEYAbGuDiJ8swPof42
    5JjnIJ70dQv4wF1qHfu+ya+LPNnRZW2gF6Hbj8ECgYB+/Yu7Xnl9rIbDUNWWG3YS
    as5zej+7DEUs1aOIKHsvAcGEWK/O+/dDK61cnHA30MpcQ3jsW2FlCvmThfRUbTKc
    7JqGsJrTeswCEhCal5EmW1SOB3KDBq6m8MMHbjzx+W7/M5Cn0s5U9scoMn/ITi5u
    hDNC+Z1yYcPUwj89VvyuVQKBgCq9ztxC3vyp7Gf9xJsJ9oG3XfzeKrm2HcBUf2vC
    8txhZWmWpQIeDhRH+i8OvWCwOodCFrxGZ6dWqqX4f/9X4BvFXy/Dv80Ldb6X9O98
    ocYKZ9Rl+wiUbQGuLQd8eTlsoBOMfkDrM6U6pkYzMiLDo2b3nN9DTKVIDbv5Q+tr
    YnbBAoGBAMSqyBvyAsNdek5qvQ5yNqOl6X0HZJ/hp9HYz8gLyqmLWkQTjHUBI9zA
    KwJD5aLsPQCvKrE1oW66XkqI0p1KqZtkfL9ZfFH9A/AhpmEqtTStT62q/Ea21MDU
    AqJIPMXO96oTVY6eorI9BU0cG6n0UvzWZAxNT2UDK07UTB4v7Z8y
    -----END RSA PRIVATE KEY-----
""")
    config.CONFIG.context = []
    self.assertIsInstance(
        config.CONFIG.Get("PrivateKeys.executable_signing_private_key"),
        rdf_crypto.RSAPrivateKey,
    )


class CryptoUtilTest(CryptoTestBase):

  def testStreamingCBCEncryptor(self):
    key = rdf_crypto.EncryptionKey.GenerateKey()
    iv = rdf_crypto.EncryptionKey.GenerateKey()
    # 160 characters.
    message = b"Hello World!!!!!" * 10

    for plaintext, partitions in [
        (
            message,
            [
                [160],
                [80, 80],
                [75, 75, 10],
                [1, 159],
                [10] * 16,
                [1] * 160,
            ],
        ),
        # Prime length, not a multiple of blocksize.
        (
            message[:149],
            [
                [149],
                [80, 69],
                [75, 55, 19],
                [1, 148],
                [10] * 14 + [9],
                [1] * 149,
            ],
        ),
    ]:
      for partition in partitions:
        cipher = rdf_crypto.AES128CBCCipher(key, iv)
        streaming_cbc = rdf_crypto.StreamingCBCEncryptor(cipher)
        offset = 0
        out = []
        for n in partition:
          next_partition = plaintext[offset : offset + n]
          out.append(streaming_cbc.Update(next_partition))
          offset += n
        out.append(streaming_cbc.Finalize())

        self.assertEqual(cipher.Decrypt(b"".join(out)), plaintext)

  def testEncryptionKey(self):
    key = rdf_crypto.EncryptionKey.GenerateKey()
    iv = rdf_crypto.EncryptionKey.GenerateKey()

    self.assertNotEqual(key, iv)
    self.assertNotEqual(key.RawBytes(), iv.RawBytes())

    # This key is too short.
    self.assertRaises(rdf_crypto.CipherError, rdf_crypto.EncryptionKey, b"foo")

    copied_key = rdf_crypto.EncryptionKey(key.RawBytes())
    self.assertEqual(copied_key, key)
    self.assertEqual(copied_key.RawBytes(), key.RawBytes())

  def testAES128CBCCipher(self):
    key = rdf_crypto.EncryptionKey.GenerateKey()
    iv = rdf_crypto.EncryptionKey.GenerateKey()

    cipher = rdf_crypto.AES128CBCCipher(key, iv)

    plain_text = b"hello world!"
    cipher_text = cipher.Encrypt(plain_text)

    # Repeatedly calling Encrypt should repeat the same cipher text.
    self.assertEqual(cipher_text, cipher.Encrypt(plain_text))

    self.assertNotEqual(cipher_text, plain_text)
    self.assertEqual(cipher.Decrypt(cipher_text), plain_text)

    key2 = rdf_crypto.EncryptionKey.GenerateKey()
    iv2 = rdf_crypto.EncryptionKey.GenerateKey()
    cipher = rdf_crypto.AES128CBCCipher(key, iv2)
    self.assertRaises(rdf_crypto.CipherError, cipher.Decrypt, plain_text)
    cipher = rdf_crypto.AES128CBCCipher(key2, iv)
    self.assertRaises(rdf_crypto.CipherError, cipher.Decrypt, plain_text)
    cipher = rdf_crypto.AES128CBCCipher(key2, iv2)
    self.assertRaises(rdf_crypto.CipherError, cipher.Decrypt, plain_text)


class SymmetricCipherTest(
    rdf_test_base.RDFValueTestMixin, test_lib.GRRBaseTest
):
  rdfvalue_class = rdf_crypto.SymmetricCipher

  sample_cache = {}

  def GenerateSample(self, seed=1):
    # We need to generate consistent new samples for each seed.
    result = SymmetricCipherTest.sample_cache.get(seed)
    if result is None:
      result = self.rdfvalue_class.Generate("AES128CBC")
      SymmetricCipherTest.sample_cache[seed] = result

    return result

  def _testEncrypt(self, plain_text):
    sample = self.GenerateSample()
    self.assertLen(sample._key.RawBytes(), 16)
    self.assertLen(sample._iv.RawBytes(), 16)
    self.assertEqual(sample._key.RawBytes(), sample._key)

    cipher_text = sample.Encrypt(plain_text)
    self.assertNotEqual(cipher_text, plain_text)
    self.assertEqual(sample.Decrypt(cipher_text), plain_text)

  def testEncrypt(self):
    self._testEncrypt(b"hello world!")

  def testLargeEncrypt(self):
    # Test with a plaintext that is longer than blocksize.
    self._testEncrypt(b"hello world!" * 100)


class RSATest(CryptoTestBase):

  def testPassPhraseEncryption(self):
    passphrase = b"testtest"
    key = rdf_crypto.RSAPrivateKey.GenerateKey()
    protected_pem = key.AsPassphraseProtectedPEM(passphrase)
    unprotected_pem = key.AsPEM()

    with mock.patch.object(utils, "PassphraseCallback", lambda: passphrase):

      # Key from unprotected PEM should always work.
      rdf_crypto.RSAPrivateKey(unprotected_pem, allow_prompt=False)

      # Protected PEM does not work if we don't allow prompts.
      with self.assertRaises(type_info.TypeValueError):
        rdf_crypto.RSAPrivateKey(protected_pem, allow_prompt=False)

      # If we allow prompts, this will work.
      rdf_crypto.RSAPrivateKey(protected_pem, allow_prompt=True)

      # Default is to not ask unless we are in a command line context.
      with self.assertRaises(type_info.TypeValueError):
        rdf_crypto.RSAPrivateKey(protected_pem)

      with mock.patch.object(
          config.CONFIG,
          "context",
          config.CONFIG.context + ["Commandline Context"],
      ):
        rdf_crypto.RSAPrivateKey(protected_pem)

        # allow_prompt=False even prevents this in the Commandline Context.
        with self.assertRaises(type_info.TypeValueError):
          rdf_crypto.RSAPrivateKey(protected_pem, allow_prompt=False)

  def testSignVerify(self):

    private_key = rdf_crypto.RSAPrivateKey.GenerateKey(bits=2048)
    public_key = private_key.GetPublicKey()

    message = b"Hello World!"

    signature = private_key.Sign(message)

    # If this fails, it raises.
    public_key.Verify(message, signature)

    # Make sure it does.
    broken_signature = _Tamper(signature)
    broken_message = _Tamper(message)

    self.assertRaises(
        rdf_crypto.VerificationError,
        public_key.Verify,
        message,
        broken_signature,
    )
    self.assertRaises(
        rdf_crypto.VerificationError,
        public_key.Verify,
        broken_message,
        signature,
    )
    self.assertRaises(
        rdf_crypto.VerificationError, public_key.Verify, message, b""
    )

  def testEncryptDecrypt(self):
    private_key = rdf_crypto.RSAPrivateKey.GenerateKey(bits=2048)
    public_key = private_key.GetPublicKey()

    message = b"Hello World!"

    ciphertext = public_key.Encrypt(message)
    self.assertNotEqual(ciphertext, message)

    plaintext = private_key.Decrypt(ciphertext)
    self.assertEqual(plaintext, message)

    self.assertRaises(
        rdf_crypto.CipherError, private_key.Decrypt, _Tamper(ciphertext)
    )

  def testM2CryptoSigningCompatibility(self):
    pem = open(os.path.join(self.base_path, "m2crypto/rsa_key"), "rb").read()
    signature = open(
        os.path.join(self.base_path, "m2crypto/signature"), "rb"
    ).read()
    private_key = rdf_crypto.RSAPrivateKey(pem)
    message = b"Signed by M2Crypto!"

    public_key = private_key.GetPublicKey()

    # If this doesn't raise InvalidSignature, we are good.
    public_key.Verify(message, signature)

  def testM2CryptoEncryptionCompatibility(self):
    pem = open(os.path.join(self.base_path, "m2crypto/rsa_key"), "rb").read()
    private_key = rdf_crypto.RSAPrivateKey(pem)
    ciphertext = open(
        os.path.join(self.base_path, "m2crypto/rsa_ciphertext"), "rb"
    ).read()
    message = b"Encrypted by M2Crypto!"

    plaintext = private_key.Decrypt(ciphertext)
    self.assertEqual(plaintext, message)


class HMACTest(CryptoTestBase):

  def testHMAC(self):
    """A basic test for the HMAC class."""
    key = rdf_crypto.EncryptionKey.GenerateKey()
    message = b"Hello World!"
    h = rdf_crypto.HMAC(key)
    signature = h.HMAC(message)

    h.Verify(message, signature)

    broken_message = message + b"!"
    self.assertRaises(
        rdf_crypto.VerificationError, h.Verify, broken_message, signature
    )

    broken_signature = _Tamper(signature)
    self.assertRaises(
        rdf_crypto.VerificationError,
        h.Verify,
        b"Hello World!",
        broken_signature,
    )

  def testSHA256(self):
    """Tests that both types of signatures are ok."""
    key = rdf_crypto.EncryptionKey.GenerateKey()
    message = b"Hello World!"
    h = rdf_crypto.HMAC(key)
    signature_sha1 = h.HMAC(message)
    signature_sha256 = h.HMAC(message, use_sha256=True)

    self.assertNotEqual(signature_sha1, signature_sha256)
    h.Verify(message, signature_sha1)
    h.Verify(message, signature_sha256)

  def testM2CryptoCompatibility(self):
    message = b"HMAC by M2Crypto!"
    signature = binascii.unhexlify("99cae3ec7b41ceb6e6619f2f85368cb3ae118b70")
    key = rdf_crypto.EncryptionKey.FromHumanReadable(
        "94bd4e0ecc8397a8b2cdbc4b127ee7b0"
    )
    h = rdf_crypto.HMAC(key)

    self.assertEqual(h.HMAC(message), signature)

    h.Verify(message, signature)


class RDFX509CertTest(CryptoTestBase):

  def testExpiredTestCertificate(self):
    pem = open(
        os.path.join(self.base_path, "outdated_certificate"), "rb"
    ).read()
    certificate = rdf_crypto.RDFX509Cert(pem)

    exception_catcher = self.assertRaises(rdf_crypto.VerificationError)
    with exception_catcher:
      # We don't pass a proper key here, this will fail before it even touches
      # the key.
      certificate.Verify(None)

    self.assertIn("Certificate expired!", str(exception_catcher.exception))


class PasswordTest(CryptoTestBase):

  def testPassword(self):
    sample = jobs_pb2.Password()

    rdf_crypto.SetPassword(sample, "foo")
    serialized = sample.SerializeToString()
    self.assertNotIn(b"foo", serialized)

    read_sample = jobs_pb2.Password()
    read_sample.ParseFromString(serialized)

    self.assertFalse(rdf_crypto.CheckPassword(sample, "bar"))
    self.assertFalse(rdf_crypto.CheckPassword(read_sample, "bar"))
    self.assertTrue(rdf_crypto.CheckPassword(sample, "foo"))
    self.assertTrue(rdf_crypto.CheckPassword(read_sample, "foo"))


def _Tamper(string):
  return string[:-1] + bytes([string[-1] ^ 1])


def main(argv):
  # Run the full test suite
  test_lib.main(argv)


if __name__ == "__main__":
  app.run(main)



#!/usr/bin/env python
"""Filesystem related client rdfvalues."""

import stat

from grr_response_core.lib import rdfvalue
from grr_response_core.lib.rdfvalues import client_action as rdf_client_action
from grr_response_core.lib.rdfvalues import paths as rdf_paths
from grr_response_core.lib.rdfvalues import protodict as rdf_protodict
from grr_response_core.lib.rdfvalues import standard as rdf_standard
from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_proto import flows_pb2
from grr_response_proto import jobs_pb2
from grr_response_proto import sysinfo_pb2


# These are objects we store as attributes of the client.
class Filesystem(rdf_structs.RDFProtoStruct):
  """A filesystem on the client.

  This class describes a filesystem mounted on the client.
  """

  protobuf = sysinfo_pb2.Filesystem
  rdf_deps = [
      rdf_protodict.AttributedDict,
  ]


class WindowsVolume(rdf_structs.RDFProtoStruct):
  """A disk volume on a windows client."""

  protobuf = sysinfo_pb2.WindowsVolume


class UnixVolume(rdf_structs.RDFProtoStruct):
  """A disk volume on a unix client."""

  protobuf = sysinfo_pb2.UnixVolume


class Volume(rdf_structs.RDFProtoStruct):
  """A disk volume on the client."""

  protobuf = sysinfo_pb2.Volume
  rdf_deps = [
      rdfvalue.RDFDatetime,
      UnixVolume,
      WindowsVolume,
  ]

  def FreeSpacePercent(self):
    try:
      return (
          self.actual_available_allocation_units / self.total_allocation_units
      ) * 100.0
    except ZeroDivisionError:
      return 100

  def FreeSpaceBytes(self):
    return self.AUToBytes(self.actual_available_allocation_units)

  def AUToBytes(self, allocation_units):
    """Convert a number of allocation units to bytes."""
    return (
        allocation_units
        * self.sectors_per_allocation_unit
        * self.bytes_per_sector
    )

  def AUToGBytes(self, allocation_units):
    """Convert a number of allocation units to GigaBytes."""
    return self.AUToBytes(allocation_units) // 1000.0**3

  def Name(self):
    """Return the best available name for this volume."""
    return (
        self.name
        or self.device_path
        or self.windowsvolume.drive_letter
        or self.unixvolume.mount_point
        or None
    )


class DiskUsage(rdf_structs.RDFProtoStruct):
  protobuf = sysinfo_pb2.DiskUsage


class StatMode(rdfvalue.RDFInteger):
  """The mode of a file."""

  protobuf_type = "unsigned_integer"

  def __str__(self) -> str:
    """Pretty print the file mode."""
    type_char = "-"

    mode = int(self)
    if stat.S_ISREG(mode):
      type_char = "-"
    elif stat.S_ISBLK(mode):
      type_char = "b"
    elif stat.S_ISCHR(mode):
      type_char = "c"
    elif stat.S_ISDIR(mode):
      type_char = "d"
    elif stat.S_ISFIFO(mode):
      type_char = "p"
    elif stat.S_ISLNK(mode):
      type_char = "l"
    elif stat.S_ISSOCK(mode):
      type_char = "s"

    mode_template = "rwx" * 3
    # Strip the "0b"
    bin_mode = bin(int(self))[2:]
    bin_mode = bin_mode[-9:]
    bin_mode = "0" * (9 - len(bin_mode)) + bin_mode

    bits = []
    for i, mode_ in enumerate(mode_template):
      if bin_mode[i] == "1":
        bit = mode_
      else:
        bit = "-"

      bits.append(bit)

    if stat.S_ISUID & mode:
      bits[2] = "S"
    if stat.S_ISGID & mode:
      bits[5] = "S"
    if stat.S_ISVTX & mode:
      if bits[8] == "x":
        bits[8] = "t"
      else:
        bits[8] = "T"

    return type_char + "".join(bits)


class StatExtFlagsOsx(rdfvalue.RDFInteger):
  """Extended file attributes for Mac (set by `chflags`)."""

  protobuf_type = "unsigned_integer_32"


class StatExtFlagsLinux(rdfvalue.RDFInteger):
  """Extended file attributes as reported by `lsattr`."""

  protobuf_type = "unsigned_integer_32"


class ExtAttr(rdf_structs.RDFProtoStruct):
  """An RDF value representing an extended attributes of a file."""

  protobuf = jobs_pb2.StatEntry.ExtAttr


class StatEntry(rdf_structs.RDFProtoStruct):
  """Represent an extended stat response."""

  protobuf = jobs_pb2.StatEntry
  rdf_deps = [
      rdf_protodict.DataBlob,
      rdf_paths.PathSpec,
      rdfvalue.RDFDatetimeSeconds,
      StatMode,
      StatExtFlagsOsx,
      StatExtFlagsLinux,
      ExtAttr,
  ]

  def AFF4Path(self, client_urn):
    return self.pathspec.AFF4Path(client_urn)


class FindSpec(rdf_structs.RDFProtoStruct):
  """A find specification."""

  protobuf = jobs_pb2.FindSpec
  rdf_deps = [
      rdfvalue.RDFBytes,
      rdf_paths.GlobExpression,
      rdf_client_action.Iterator,
      rdf_paths.PathSpec,
      rdfvalue.RDFDatetime,
      rdf_standard.RegularExpression,
      StatEntry,
      StatMode,
  ]

  def Validate(self):
    """Ensure the pathspec is valid."""
    self.pathspec.Validate()

    if (
        self.HasField("start_time")
        and self.HasField("end_time")
        and self.start_time > self.end_time
    ):
      raise ValueError("Start time must be before end time.")

    if not self.path_regex and not self.data_regex and not self.path_glob:
      raise ValueError(
          "A Find specification can not contain both an empty "
          "path regex and an empty data regex"
      )


class BareGrepSpec(rdf_structs.RDFProtoStruct):
  """A GrepSpec without a target."""

  protobuf = flows_pb2.BareGrepSpec
  rdf_deps = [
      rdf_standard.LiteralExpression,
      rdf_standard.RegularExpression,
  ]


class GrepSpec(rdf_structs.RDFProtoStruct):
  protobuf = jobs_pb2.GrepSpec
  rdf_deps = [
      rdfvalue.RDFBytes,
      rdf_standard.LiteralExpression,
      rdf_paths.PathSpec,
  ]

  def Validate(self):
    self.target.Validate()


class BlobImageChunkDescriptor(rdf_structs.RDFProtoStruct):
  """A descriptor of a file chunk stored in VFS blob image."""

  protobuf = jobs_pb2.BlobImageChunkDescriptor
  rdf_deps = []


class BlobImageDescriptor(rdf_structs.RDFProtoStruct):
  """A descriptor of a file stored as VFS blob image."""

  protobuf = jobs_pb2.BlobImageDescriptor
  rdf_deps = [BlobImageChunkDescriptor]



#!/usr/bin/env python
"""Local contexts.

Put any config.CONFIG.DEFINE_context statements specific to your deployment
in this file.
"""

# pylint: disable=unused-import
from grr_response_core.lib import config_lib



#!/usr/bin/env python
"""This directory contains local site-specific artifacts.

All .yaml files in this directory will be automatically loaded by the artifact
library (See the configuration parameter Artifacts.artifact_dirs.
"""



#!/usr/bin/env python
"""RDFValues for GRR client-side cron jobs parsing."""

from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_proto import sysinfo_pb2


class CronTabEntry(rdf_structs.RDFProtoStruct):
  protobuf = sysinfo_pb2.CronTabEntry


class CronTabFile(rdf_structs.RDFProtoStruct):
  protobuf = sysinfo_pb2.CronTabFile
  rdf_deps = [
      CronTabEntry,
  ]



#!/usr/bin/env python
"""DO NOT USE. Deprecated RDFValues."""

from grr_response_core.lib import rdfvalue
from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_proto import deprecated_pb2


class CpuSample(rdf_structs.RDFProtoStruct):
  """A single CPU sample."""

  protobuf = deprecated_pb2.CpuSample
  rdf_deps = [
      rdfvalue.RDFDatetime,
  ]


class IOSample(rdf_structs.RDFProtoStruct):
  """A single I/O sample as collected by `psutil`."""

  protobuf = deprecated_pb2.IOSample
  rdf_deps = [
      rdfvalue.RDFDatetime,
  ]


class ClientStats(rdf_structs.RDFProtoStruct):
  """A client stat object."""

  protobuf = deprecated_pb2.ClientStats
  rdf_deps = [
      CpuSample,
      IOSample,
      rdfvalue.RDFDatetime,
  ]



#!/usr/bin/env python
"""Implementation of configuration_file types."""

from grr_response_core.lib.rdfvalues import protodict as rdf_protodict
from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_proto import config_file_pb2


class LogTarget(rdf_structs.RDFProtoStruct):
  """An RDFValue representation of a logging target."""

  protobuf = config_file_pb2.LogTarget


class LogConfig(rdf_structs.RDFProtoStruct):
  """An RDFValue representation of a logging configuration."""

  protobuf = config_file_pb2.LogConfig
  rdf_deps = [
      LogTarget,
  ]


class NfsClient(rdf_structs.RDFProtoStruct):
  """An RDFValue representation of an NFS Client configuration."""

  protobuf = config_file_pb2.NfsClient


class NfsExport(rdf_structs.RDFProtoStruct):
  """An RDFValue representation of an NFS Export entry."""

  protobuf = config_file_pb2.NfsExport
  rdf_deps = [
      NfsClient,
  ]


class NtpConfig(rdf_structs.RDFProtoStruct):
  """An RDFValue representation of a ntp config file."""

  protobuf = config_file_pb2.NtpConfig
  rdf_deps = [
      rdf_protodict.AttributedDict,
  ]


class PamConfigEntry(rdf_structs.RDFProtoStruct):
  """An RDFValue representation of a single entry in a PAM configuration."""

  protobuf = config_file_pb2.PamConfigEntry


class PamConfig(rdf_structs.RDFProtoStruct):
  """An RDFValue representation of an entire PAM configuration."""

  protobuf = config_file_pb2.PamConfig
  rdf_deps = [
      PamConfigEntry,
  ]



#!/usr/bin/env python
import csv
import io

from absl.testing import absltest

from grr_response_core.lib.util import body
from grr_response_proto import timeline_pb2


class StreamTest(absltest.TestCase):

  def testSingle(self):
    entry = timeline_pb2.TimelineEntry()
    entry.path = "/foo/bar/baz".encode("utf-8")
    entry.mode = 0o100644
    entry.size = 42
    entry.dev = 101
    entry.ino = 404
    entry.uid = 13
    entry.gid = 7
    entry.atime_ns = 123 * 1_000_000_000
    entry.mtime_ns = 456 * 1_000_000_000
    entry.ctime_ns = 789 * 1_000_000_000
    entry.btime_ns = 1337 * 1_000_000_000

    content = b"".join(body.Stream(iter([entry]))).decode("utf-8")
    expected = "0|/foo/bar/baz|404|-rw-r--r--|13|7|42|123|456|789|1337\n"

    self.assertEqual(content, expected)

  def testMultiple(self):
    entries = []

    for idx in range(100):
      entry = timeline_pb2.TimelineEntry()
      entry.path = "/foo/bar/baz{}".format(idx).encode("utf-8")
      entry.size = idx

      entries.append(entry)

    content = b"".join(body.Stream(iter(entries))).decode("utf-8")
    reader = csv.reader(io.StringIO(content), delimiter="|")

    rows = list(reader)
    self.assertLen(rows, len(entries))

    for idx, row in enumerate(rows):
      self.assertEqual(row[1].encode("utf-8"), entries[idx].path)
      self.assertEqual(int(row[6]), entries[idx].size)

  def testChunks(self):
    entries = []
    for idx in range(1024):
      entry = timeline_pb2.TimelineEntry()
      entry.path = "/foo/bar{}".format(idx).encode("utf-8")

      entries.append(entry)

    opts = body.Opts()
    opts.chunk_size = 6

    chunks = list(body.Stream(iter(entries), opts=opts))
    self.assertLen(chunks, len(entries))

    content = b"".join(chunks).decode("utf-8")
    reader = csv.reader(io.StringIO(content), delimiter="|")

    rows = list(reader)
    self.assertLen(rows, len(entries))

    for idx, row in enumerate(rows):
      self.assertEqual(row[1].encode("utf-8"), entries[idx].path)

  def testUnicode(self):
    entry = timeline_pb2.TimelineEntry()
    entry.path = "/zażółć/gęślą/jaźń/💪".encode("utf-8")

    content = b"".join(body.Stream(iter([entry]))).decode("utf-8")
    reader = csv.reader(io.StringIO(content), delimiter="|")

    rows = list(reader)
    self.assertLen(rows, 1)
    self.assertEqual(rows[0][1].encode("utf-8"), entry.path)

  def testNonUnicode(self):
    entry = timeline_pb2.TimelineEntry()
    entry.path = b"/\xff\x00\x00/\xfb\xfa\xf5"

    content = b"".join(body.Stream(iter([entry])))
    self.assertIn(b"/\xff\x00\x00/\xfb\xfa\xf5", content)

  def testSubsecondPrecision(self):
    entry = timeline_pb2.TimelineEntry()
    entry.path = "/foo/bar".encode("utf-8")
    entry.atime_ns = 123_456_789_000

    opts = body.Opts()
    opts.timestamp_subsecond_precision = True

    stream = body.Stream(iter([entry]), opts=opts)
    content = b"".join(stream).decode("utf-8")

    self.assertIn("/foo/bar", content)
    self.assertIn("123.456789", content)

  def testNtfsFileReference(self):
    entry = timeline_pb2.TimelineEntry()
    entry.path = "/foo/bar".encode("utf-8")
    entry.ino = 1688849860339456

    opts = body.Opts()
    opts.inode_format = body.Opts.InodeFormat.NTFS_FILE_REFERENCE

    stream = body.Stream(iter([entry]), opts=opts)
    content = b"".join(stream).decode("utf-8")

    self.assertIn("/foo/bar", content)
    self.assertIn("75520-6", content)

  def testBackslashEscape(self):
    entry = timeline_pb2.TimelineEntry()
    entry.path = "C:\\Windows\\system32\\notepad.exe".encode("utf-8")

    opts = body.Opts()
    opts.backslash_escape = True

    stream = body.Stream(iter([entry]), opts=opts)
    content = b"".join(stream).decode("utf-8")

    self.assertIn("|C:\\\\Windows\\\\system32\\\\notepad.exe|", content)

  def testCarriageReturnEscape(self):
    entry = timeline_pb2.TimelineEntry()
    entry.path = "C:\\Foo\rBar\\Baz\r\rQuux".encode("utf-8")

    opts = body.Opts()
    opts.carriage_return_escape = True

    stream = body.Stream(iter([entry]), opts=opts)
    content = b"".join(stream).decode("utf-8")

    self.assertIn("|C:\\Foo\\rBar\\Baz\\r\\rQuux|", content)

  def testNonPrintableEscape(self):
    entry = timeline_pb2.TimelineEntry()
    entry.path = b"/f\x00b\x0ar\x1baz"

    opts = body.Opts()
    opts.non_printable_escape = True

    stream = body.Stream(iter([entry]), opts=opts)
    content = b"".join(stream).decode("utf-8")

    self.assertIn(r"|/f\x00b\x0ar\x1baz|", content)

  def testPathWithPipe(self):
    entry = timeline_pb2.TimelineEntry()
    entry.path = "/foo|bar/baz".encode("utf-8")

    stream = body.Stream(iter([entry]))
    content = b"".join(stream).decode("utf-8")

    self.assertIn("|/foo\\|bar/baz|", content)

  def testPathWithWhitespace(self):
    entry = timeline_pb2.TimelineEntry()
    entry.path = "/foo bar\tbaz\rquux   norf/thud".encode("utf-8")

    stream = body.Stream(iter([entry]))
    content = b"".join(stream).decode("utf-8")

    self.assertIn("|/foo bar\tbaz\rquux   norf/thud|", content)

  def testPathWithNewline(self):
    entry = timeline_pb2.TimelineEntry()
    entry.path = "/foo bar\nbaz\n\nquux/thud".encode("utf-8")

    stream = body.Stream(iter([entry]))
    content = b"".join(stream).decode("utf-8")

    self.assertIn("|/foo bar\\nbaz\\n\\nquux/thud|", content)

  def testPathWithQuote(self):
    entry = timeline_pb2.TimelineEntry()
    entry.path = '/foo"bar'.encode("utf-8")

    stream = body.Stream(iter([entry]))
    content = b"".join(stream).decode("utf-8")

    self.assertIn('|/foo"bar|', content)

  def testPathWithBackslash(self):
    entry = timeline_pb2.TimelineEntry()
    entry.path = "C:\\Windows\\system32".encode("utf-8")

    stream = body.Stream(iter([entry]))
    content = b"".join(stream).decode("utf-8")

    self.assertIn("|C:\\Windows\\system32|", content)

  def testPathWithCarriageReturn(self):
    entry = timeline_pb2.TimelineEntry()
    entry.path = "/foo/bar\rbaz/quux\r\rnorf".encode("utf-8")

    stream = body.Stream(iter([entry]))
    content = b"".join(stream).decode("utf-8")

    self.assertIn("|/foo/bar\rbaz/quux\r\rnorf|", content)

  def testModeWindows(self):
    entry_system32 = timeline_pb2.TimelineEntry()
    entry_system32.path = "C:\\Windows\\system32".encode("utf-8")
    entry_system32.attributes = 0x30

    entry_notepad = timeline_pb2.TimelineEntry()
    entry_notepad.path = "C:\\Windows\\system32\\notepad.exe".encode("utf-8")
    entry_notepad.attributes = 0x20

    entry_desktop = timeline_pb2.TimelineEntry()
    entry_desktop.path = "C:\\Users\\Public\\Desktop".encode("utf-8")
    entry_desktop.attributes = 0x13

    stream = body.Stream(iter([entry_system32, entry_notepad, entry_desktop]))
    content = b"".join(stream).decode("utf-8")

    self.assertIn("0|C:\\Windows\\system32|0|drwxrwxrwx|", content)
    self.assertIn("0|C:\\Windows\\system32\\notepad.exe|0|-rwxrwxrwx|", content)
    self.assertIn("0|C:\\Users\\Public\\Desktop|0|dr-xr-xr-x|", content)


if __name__ == "__main__":
  absltest.main()



#!/usr/bin/env python
"""These are tests for the PathSpec implementation."""

from absl import app

from grr_response_core.lib.rdfvalues import client as rdf_client
from grr_response_core.lib.rdfvalues import paths as rdf_paths
from grr_response_core.lib.rdfvalues import test_base as rdf_test_base
from grr_response_proto import jobs_pb2
from grr_response_proto import knowledge_base_pb2
from grr.test_lib import test_lib


class PathSpecTest(rdf_test_base.RDFProtoTestMixin, test_lib.GRRBaseTest):
  """Test the PathSpec implementation."""

  rdfvalue_class = rdf_paths.PathSpec

  def CheckRDFValue(self, rdfproto, sample):
    """Check that the rdfproto is the same as the sample."""
    super().CheckRDFValue(rdfproto, sample)

    self.assertEqual(rdfproto.path, sample.path)
    self.assertEqual(rdfproto.pathtype, sample.pathtype)

  def GenerateSample(self, number=0):
    """Make a sample PathSpec instance."""
    return rdf_paths.PathSpec(path="/%s/" % number, pathtype=number)

  def testPop(self):
    """Test we can pop arbitrary elements from the pathspec."""
    sample = rdf_paths.PathSpec(
        path="/", pathtype=rdf_paths.PathSpec.PathType.OS
    )

    for i in range(5):
      sample.Append(path=str(i), pathtype=rdf_paths.PathSpec.PathType.OS)

    self.assertEqual([x.path for x in sample], list("/01234"))

    # Check we pop the right element.
    popped = sample.Pop(2)
    self.assertIsInstance(popped, rdf_paths.PathSpec)
    self.assertEqual(popped.path, "1")
    self.assertEqual([x.path for x in sample], list("/0234"))

    # The first element needs special treatment.
    self.assertEqual(sample.Pop(0).path, "/")
    self.assertEqual([x.path for x in sample], list("0234"))

  def testPathSpec(self):
    """Test that PathSpec works."""
    # Make a template pathspec using a protobuf the hard way.
    pathspec_pb = jobs_pb2.PathSpec(path="/", pathtype=1)
    pathspec_pb.nested_path.path = "foo"
    pathspec_pb.nested_path.pathtype = 2

    reference_pathspec = rdf_paths.PathSpec.FromSerializedBytes(
        pathspec_pb.SerializeToString()
    )

    # Create a new RDFPathspec from scratch.
    pathspec = rdf_paths.PathSpec()
    pathspec.path = "/"
    pathspec.pathtype = 1
    pathspec.Append(path="foo", pathtype=2)

    self.assertEqual(pathspec, reference_pathspec)

    # Create a new RDFPathspec from keywords.
    pathspec = rdf_paths.PathSpec(path="/", pathtype=1)
    pathspec.Append(path="foo", pathtype=2)

    self.assertEqual(pathspec, reference_pathspec)

    # Check that copies are ok
    pathspec = pathspec.Copy()

    self.assertEqual(pathspec, reference_pathspec)

    # Accessors:
    self.assertEqual(pathspec.path, "/")
    self.assertEqual(pathspec.last.path, "foo")

    pathspec.first.path = "test"
    self.assertEqual(pathspec.last.path, "foo")

    # Test Pathspec iterator.
    self.assertEqual([x.path for x in pathspec], ["test", "foo"])

    # Length.
    self.assertLen(pathspec, 2)

    pathspec = rdf_paths.PathSpec(path="/foo", pathtype=1)
    pathspec.Append(path="/", pathtype=0)
    self.assertEqual(pathspec.Dirname().CollapsePath(), "/")
    pathspec.Append(path="sdasda", pathtype=0)
    self.assertEqual(pathspec.Dirname().CollapsePath(), "/foo")

    pathspec = rdf_paths.PathSpec(path="/foo", pathtype=1)
    pathspec_base = rdf_paths.PathSpec()
    pathspec_base.Append(pathspec)

    self.assertEqual(pathspec_base.CollapsePath(), "/foo")

    pathspec_base = rdf_paths.PathSpec()
    pathspec_base.Insert(0, path="/foo", pathtype=1)

    self.assertEqual(pathspec_base.CollapsePath(), "/foo")

  def testUnicodePaths(self):
    """Test that we can manipulate paths in unicode."""
    sample = rdf_paths.PathSpec(pathtype=1, path="/dev/c/msn升级程序[1].exe")

    # Ensure we can convert to a string.
    str(sample)

  def testCopy(self):
    sample = rdf_paths.PathSpec(
        path="/", pathtype=rdf_paths.PathSpec.PathType.OS
    )
    sample.Append(path="foo", pathtype=rdf_paths.PathSpec.PathType.TSK)

    # Make a copy of the original and change it.
    sample_copy = sample.Copy()
    sample_copy.last.path = "bar"

    # This should not change the original.
    self.assertEqual(sample.last.path, "foo")

  def testOsConstructor(self):
    pathspec = rdf_paths.PathSpec.OS(path="foo")

    self.assertEqual(pathspec.path, "foo")
    self.assertEqual(pathspec.pathtype, rdf_paths.PathSpec.PathType.OS)

  def testTskConstructor(self):
    pathspec = rdf_paths.PathSpec.TSK(mount_point="C:\\")

    self.assertEqual(pathspec.mount_point, "C:\\")
    self.assertEqual(pathspec.pathtype, rdf_paths.PathSpec.PathType.TSK)

  def testNtfsConstructor(self):
    pathspec = rdf_paths.PathSpec.NTFS(mount_point="C:\\")

    self.assertEqual(pathspec.mount_point, "C:\\")
    self.assertEqual(pathspec.pathtype, rdf_paths.PathSpec.PathType.NTFS)

  def testRegistryConstructor(self):
    pathspec = rdf_paths.PathSpec.Registry(path="HKLM\\System\\foo\\bar")

    self.assertEqual(pathspec.path, "HKLM\\System\\foo\\bar")
    self.assertEqual(pathspec.pathtype, rdf_paths.PathSpec.PathType.REGISTRY)

  def testTempConstructor(self):
    pathspec = rdf_paths.PathSpec.Temp(is_virtualroot=True)

    self.assertTrue(pathspec.is_virtualroot)
    self.assertEqual(pathspec.pathtype, rdf_paths.PathSpec.PathType.TMPFILE)


class GlobExpressionTest(rdf_test_base.RDFValueTestMixin, test_lib.GRRBaseTest):
  rdfvalue_class = rdf_paths.GlobExpression

  def GenerateSample(self, number=0):
    return self.rdfvalue_class("/home/%%users.username%%/*{}".format(number))

  def testGroupingInterpolation(self):
    glob_expression = rdf_paths.GlobExpression()

    interpolated = glob_expression.InterpolateGrouping("/home/*.{sh,deb}")
    self.assertCountEqual(interpolated, ["/home/*.deb", "/home/*.sh"])
    interpolated = glob_expression.InterpolateGrouping("/home/*.{sh, deb}")
    self.assertCountEqual(interpolated, ["/home/*. deb", "/home/*.sh"])
    interpolated = glob_expression.InterpolateGrouping(
        "HKEY_CLASSES_ROOT/CLSID/{16d12736-7a9e-4765-bec6-f301d679caaa}"
    )
    self.assertCountEqual(
        interpolated,
        ["HKEY_CLASSES_ROOT/CLSID/{16d12736-7a9e-4765-bec6-f301d679caaa}"],
    )

  def testValidation(self):
    glob_expression = rdf_paths.GlobExpression(
        "/home/%%users.username%%/**/.mozilla/"
    )
    glob_expression.Validate()

    glob_expression = rdf_paths.GlobExpression("/home/**/**")
    self.assertRaises(ValueError, glob_expression.Validate)

  def testRegExIsCorrectForGlobWithoutStars(self):
    glob_expression = rdf_paths.GlobExpression("/foo/bar/blah.txt")
    regex = glob_expression.AsRegEx()

    self.assertTrue(regex.Match("/foo/bar/blah.txt"))
    self.assertFalse(regex.Match("/foo/bar/blah2.txt"))
    self.assertFalse(regex.Match("/some/foo/bar/blah2.txt"))
    self.assertFalse(regex.Match("/some/foo/bar/blah2.txt/other"))

  def testRegExIsCorrectForGlobWithQuestion(self):
    glob_expression = rdf_paths.GlobExpression("/foo/bar/???.txt")
    regex = glob_expression.AsRegEx()

    self.assertTrue(regex.Match("/foo/bar/bla.txt"))
    self.assertFalse(regex.Match("/foo/bar/blah.txt"))

  def testRegExIsCorrectForGlobWithGrouping(self):
    glob_expression = rdf_paths.GlobExpression("/foo/{bar,other}/*{.txt,.exe}")
    regex = glob_expression.AsRegEx()

    self.assertTrue(regex.Match("/foo/bar/blah.txt"))
    self.assertTrue(regex.Match("/foo/other/blah2.txt"))
    self.assertTrue(regex.Match("/foo/bar/blah.exe"))
    self.assertTrue(regex.Match("/foo/other/blah2.exe"))

    self.assertFalse(regex.Match("/foo/other2/blah.txt"))
    self.assertFalse(regex.Match("/foo/bar/blah.com"))

  def testRegExIsCorrectForGlobWithSingleStar(self):
    glob_expression = rdf_paths.GlobExpression("/foo/bar/*.txt")
    regex = glob_expression.AsRegEx()

    self.assertTrue(regex.Match("/foo/bar/blah.txt"))

    self.assertFalse(regex.Match("/foo/bar/blah.plist"))
    self.assertFalse(regex.Match("/foo/bar/blah/blah.txt"))
    self.assertFalse(regex.Match("/foo/blah1/blah2/bar/blah.txt"))

  def testRegExIsCorrectForGlobWithTwoStars(self):
    glob_expression = rdf_paths.GlobExpression("/foo/**/bar.txt")
    regex = glob_expression.AsRegEx()

    self.assertTrue(regex.Match("/foo/bar.txt"))
    self.assertTrue(regex.Match("/foo/blah/bar.txt"))
    self.assertTrue(regex.Match("/foo/blah1/blah2/bar.txt"))

    self.assertFalse(regex.Match("/foo/bar.plist"))
    self.assertFalse(regex.Match("/foo/blah/bar.txt/res"))
    self.assertFalse(regex.Match("/foo/blah1/blah2/bar.txt2"))

  def testRegExIsCorrectForComplexGlob(self):
    glob_expression = rdf_paths.GlobExpression("/foo/**/bar?/*{.txt,.exe}")
    regex = glob_expression.AsRegEx()

    self.assertTrue(regex.Match("/foo/bar1/blah.txt"))
    self.assertTrue(regex.Match("/foo/bar2/blah.exe"))
    self.assertTrue(regex.Match("/foo/c1/c2/c3/bar1/blah.txt"))
    self.assertTrue(regex.Match("/foo/c1/c2/c3/bar2/blah.exe"))

    self.assertFalse(regex.Match("/foo/bar/blah.txt"))
    self.assertFalse(regex.Match("/foo/bar2/blah.com"))
    self.assertFalse(regex.Match("/foo/c1/c2/c3/bar1/blah.txt/res.txt"))
    self.assertFalse(regex.Match("/foo/c1/c2/c3/bar2/blah.exe/res.exe"))

  def testRegExIsCaseInsensitive(self):
    glob_expression = rdf_paths.GlobExpression("/foo/**/bar?/*{.txt,.exe}")
    regex = glob_expression.AsRegEx()

    self.assertTrue(regex.Match("/foo/bar1/blah.txt"))
    self.assertTrue(regex.Match("/foO/bAr1/blah.txt"))
    self.assertTrue(regex.Match("/foo/bar1/blah.TXT"))

    self.assertFalse(regex.Match("/foo/bar2/blah.com"))
    self.assertFalse(regex.Match("/foO/bAr2/blah.com"))
    self.assertFalse(regex.Match("/foo/bar2/blah.COM"))

  def testGlobExpressionSplitsIntoExplainableComponents(self):
    kb = knowledge_base_pb2.KnowledgeBase(
        users=[
            knowledge_base_pb2.User(username="foo", homedir="/home/foo"),
            knowledge_base_pb2.User(username="bar", homedir="/home/bar"),
            knowledge_base_pb2.User(username="baz", homedir="/home/baz"),
        ]
    )

    # Test for preservation of **/ because it behaves different to **.
    ge = rdf_paths.GlobExpression("/foo/**/{bar,baz}/bar?/.*baz")
    components = ge.ExplainComponents(2, kb)
    self.assertEqual(
        [c.glob_expression for c in components],
        ["/foo/", "**/", "{bar,baz}", "/bar", "?", "/.", "*", "baz"],
    )

    ge = rdf_paths.GlobExpression("/foo/**bar")
    components = ge.ExplainComponents(2, kb)
    self.assertEqual(
        [c.glob_expression for c in components], ["/foo/", "**", "bar"]
    )

    ge = rdf_paths.GlobExpression("/foo/**10bar")
    components = ge.ExplainComponents(2, kb)
    self.assertEqual(
        [c.glob_expression for c in components], ["/foo/", "**10", "bar"]
    )

    ge = rdf_paths.GlobExpression("/{foo,bar,baz}")
    components = ge.ExplainComponents(2, kb)
    self.assertEqual(components[1].examples, ["foo", "bar"])

    ge = rdf_paths.GlobExpression("%%users.homedir%%/foo")
    components = ge.ExplainComponents(2, kb)
    self.assertEqual(
        [c.glob_expression for c in components], ["%%users.homedir%%", "/foo"]
    )
    self.assertEqual(components[0].examples, ["/home/foo", "/home/bar"])

  def testExplainComponentsReturnsEmptyExamplesOnKbError(self):
    kb = knowledge_base_pb2.KnowledgeBase(users=[knowledge_base_pb2.User()])
    ge = rdf_paths.GlobExpression("%%users.appdir%%/foo")

    components = ge.ExplainComponents(2, kb)
    self.assertEqual(components[0].examples, [])

  def _testAFF4Path_mountPointResolution(
      self, pathtype: rdf_paths.PathSpec.PathType
  ) -> None:
    path = rdf_paths.PathSpec(
        path="\\\\.\\Volume{1234}\\",
        pathtype=rdf_paths.PathSpec.PathType.OS,
        mount_point="/c:/",
        nested_path=rdf_paths.PathSpec(
            path="/windows/",
            pathtype=pathtype,
        ),
    )
    prefix = rdf_paths.PathSpec.AFF4_PREFIXES[pathtype]
    self.assertEqual(
        str(path.AFF4Path(rdf_client.ClientURN("C.0000000000000001"))),
        f"aff4:/C.0000000000000001{prefix}/\\\\.\\Volume{{1234}}\\/windows",
    )

  def testAFF4Path_mountPointResolution_TSK(self):
    self._testAFF4Path_mountPointResolution(rdf_paths.PathSpec.PathType.TSK)

  def testAFF4Path_mountPointResolution_NTFS(self):
    self._testAFF4Path_mountPointResolution(rdf_paths.PathSpec.PathType.NTFS)


def main(argv):
  # Run the full test suite
  test_lib.main(argv)


if __name__ == "__main__":
  app.run(main)



#!/usr/bin/env python
"""A module with utilities for working with iterators."""

from collections.abc import Iterator
from typing import Optional, TypeVar

_T = TypeVar("_T")


class NoYieldsError(ValueError):
  """An error raised when assuming single item on empty iterators."""


class TooManyYieldsError(ValueError):
  """An error raised when assuming single item on multi-yield iterators."""


def AssumeSingle(items: Iterator[_T]) -> _T:
  """Retrieves a result from a single-yield iterator.

  Args:
    items: An iterator that is expected to yield a single value.

  Returns:
    The only value this iterator yields.

  Raises:
    NoYieldsError: If `iterator` yields no items (instead of one).
    TooManyYieldsError: If `iterator` yields many items (instead of one).
  """
  try:
    result = next(items)
  except StopIteration:
    raise NoYieldsError()

  try:
    next(items)
    raise TooManyYieldsError()
  except StopIteration:
    pass

  return result


class Counted(Iterator[_T]):
  """An iterator wrapper that counts number of iterated items."""

  def __init__(self, inner: Iterator[_T]) -> None:
    """Initializes the wrapper.

    Args:
      inner: An iterator to wrap.
    """
    super().__init__()

    self._inner: Iterator[_T] = inner
    self._count: int = 0

  def __iter__(self) -> Iterator[_T]:
    return self

  def __next__(self) -> _T:
    item = next(self._inner)
    self._count += 1

    return item

  @property
  def count(self) -> int:
    """Retrieves the number of iterated items of the wrapped iterator."""
    return self._count

  def Reset(self) -> None:
    """Resets the counter."""
    self._count = 0


class Lookahead(Iterator[_T]):
  """An iterator wrapper that allows to peek at items without consuming them."""

  def __init__(self, inner: Iterator[_T]) -> None:
    """Initializes the wrapper.

    Args:
      inner: An iterator to wrap.
    """
    super().__init__()

    self._inner: Iterator[_T] = inner

    self._item: Optional[_T] = None
    self._done: bool = False

    self._Pull()

  def __iter__(self) -> Iterator[_T]:
    return self

  def __next__(self) -> _T:
    if self.done:
      raise StopIteration()

    # Because of the `self.done` check, `result` is guaranteed to be set to some
    # value pulled from the iterator. Ideally, we would like to assert that it
    # is not `None` but this is not possible as some iterators might yield it as
    # a completely legitimate value.
    result = self.item

    self._Pull()

    return result

  @property
  def item(self) -> _T:
    """Retrieves the current item from the iterator."""
    if self.done:
      raise ValueError("No more items available")

    return self._item

  @property
  def done(self) -> bool:
    """Checks whether the iterator has already yielded all items."""
    return self._done

  def _Pull(self) -> None:
    """Pulls the next item from the iterator."""
    try:
      self._item = next(self._inner)
    except StopIteration:
      self._done = True



#!/usr/bin/env python
"""A module with RDF wrappers for Protocol Buffers Well-Known Types."""

from google.protobuf import timestamp_pb2
from grr_response_core.lib.rdfvalues import structs as rdf_structs


class Timestamp(rdf_structs.RDFProtoStruct):
  """An RDF wrapper for the `Timestamp` Well-Known Type."""

  protobuf = timestamp_pb2.Timestamp



#!/usr/bin/env python
"""AFF4 RDFValue implementations.

This module contains all RDFValue implementations.

NOTE: This module uses the class registry to contain all implementations of
RDFValue class, regardless of where they are defined. To do this reliably, these
implementations must be imported _before_ the relevant classes are referenced
from this module.
"""

import abc
import calendar
import datetime
import functools
import logging
import posixpath
import re
import time
from typing import Any, Optional, Union, cast
import zlib

import dateutil
from dateutil import parser

from google.protobuf import timestamp_pb2
from grr_response_core.lib import registry
from grr_response_core.lib import utils
from grr_response_core.lib.util import precondition
from grr_response_core.lib.util import random
from grr_response_core.lib.util import text

# Somewhere to keep all the late binding placeholders.
_LATE_BINDING_STORE = {}


def RegisterLateBindingCallback(target_name, callback, **kwargs):
  """Registers a callback to be invoked when the RDFValue named is declared."""
  _LATE_BINDING_STORE.setdefault(target_name, []).append((callback, kwargs))


class Error(Exception):
  """Errors generated by RDFValue parsers."""


class InitializeError(Error):
  """Raised when we can not initialize from this parameter."""


class DecodeError(InitializeError, ValueError):
  """Generated when we can not decode the data."""

  def __init__(self, msg):
    logging.debug(msg)
    super().__init__(msg)


class RDFValueMetaclass(registry.MetaclassRegistry):
  """A metaclass for managing semantic values."""

  def __init__(cls, name, bases, env_dict):  # pylint: disable=no-self-argument
    super(RDFValueMetaclass, cls).__init__(name, bases, env_dict)

    # Run and clear any late binding callbacks registered for this class.
    for callback, kwargs in _LATE_BINDING_STORE.pop(name, []):
      callback(target=cls, **kwargs)


class RDFValue(metaclass=RDFValueMetaclass):  # pylint: disable=invalid-metaclass
  """Baseclass for values.

  RDFValues are serialized to and from the data store.
  """

  # This is how the attribute will be serialized to the data store. It must
  # indicate both the type emitted by SerializeToWireFormat() and expected by
  # FromWireFormat()
  protobuf_type = "bytes"

  # URL pointing to a help page about this value type.
  context_help_url = None

  _value = None
  _prev_hash = None

  # Mark as dirty each time we modify this object.
  dirty = False

  # If this value was created as part of an AFF4 attribute, the attribute is
  # assigned here.
  attribute_instance = None

  def __init__(self):
    self._prev_hash = None

  def Copy(self):
    """Make a new copy of this RDFValue."""
    return self.__class__.FromSerializedBytes(self.SerializeToBytes())

  def __copy__(self):
    return self.Copy()

  @classmethod
  def FromWireFormat(cls, value):
    raise NotImplementedError(
        "Class {} does not implement FromWireFormat.".format(cls.__name__)
    )

  @classmethod
  def FromSerializedBytes(cls, value: bytes):
    raise NotImplementedError(
        "Class {} does not implement FromSerializedBytes.".format(cls.__name__)
    )

  # TODO: Remove legacy SerializeToWireFormat.
  def SerializeToWireFormat(self):
    """Serialize to a datastore compatible form."""
    return self.SerializeToBytes()

  @abc.abstractmethod
  def SerializeToBytes(self):
    """Serialize into a string which can be parsed using FromSerializedBytes."""

  @classmethod
  def Fields(cls):
    """Return a list of fields which can be queried from this value."""
    return []

  def __eq__(self, other):
    return self._value == other

  def __ne__(self, other):
    return not self.__eq__(other)

  def __hash__(self):
    new_hash = hash(self.SerializeToBytes())
    if self._prev_hash is not None and new_hash != self._prev_hash:
      raise AssertionError(
          "Usage of {} violates Python data model: hash() has changed! Usage "
          "of RDFStructs as members of sets or keys of dicts is discouraged. "
          "If used anyway, mutating is prohibited, because it causes the hash "
          "to change. Be aware that accessing unset fields can trigger a "
          "mutation.".format(type(self).__name__)
      )
    else:
      self._prev_hash = new_hash
      return new_hash

  def __bool__(self):
    return bool(self._value)

  def __str__(self):
    """Ignores the __repr__ override below to avoid indefinite recursion."""
    return super().__repr__()

  def __repr__(self):
    content = str(self)

    # Note %r, which prevents nasty nonascii characters from being printed,
    # including dangerous terminal escape sequences.
    return "<%s(%r)>" % (self.__class__.__name__, content)


RDFValue.classes["bool"] = bool
RDFValue.classes["RDFBool"] = bool


class RDFPrimitive(RDFValue):
  """An immutable RDFValue that wraps a primitive value (e.g. int)."""

  _primitive_value = None

  def __init__(self, initializer):
    super().__init__()
    self._primitive_value = initializer

  @property
  def _value(self):
    return self._primitive_value

  @classmethod
  def FromHumanReadable(cls, string: str):
    """Returns a new instance from a human-readable string.

    Args:
      string: An `unicode` value to initialize the object from.
    """
    raise NotImplementedError(
        "Class {} does not implement FromHumanReadable.".format(cls.__name__)
    )


@functools.total_ordering
class RDFBytes(RDFPrimitive):
  """An attribute which holds bytes."""

  protobuf_type = "bytes"

  def __init__(self, initializer=None):
    if initializer is None:
      super().__init__(b"")
    elif isinstance(initializer, bytes):
      super().__init__(initializer)
    elif isinstance(initializer, RDFBytes):
      super().__init__(initializer.AsBytes())
    else:
      message = "Unexpected initializer `{value}` of type {type}"
      raise TypeError(message.format(value=initializer, type=type(initializer)))

  @classmethod
  def FromSerializedBytes(cls, value: bytes):
    precondition.AssertType(value, bytes)
    return cls(value)

  @classmethod
  def FromWireFormat(cls, value: bytes):
    precondition.AssertType(value, bytes)
    return cls(value)

  @classmethod
  def FromHumanReadable(cls, string: str):
    precondition.AssertType(string, str)
    return cls(string.encode("utf-8"))

  def AsBytes(self):
    return self._value

  def SerializeToBytes(self):
    return self.AsBytes()

  def __str__(self) -> str:
    return text.Hexify(self.AsBytes())

  def __hash__(self):
    return hash(self.AsBytes())

  def __lt__(self, other):
    if isinstance(other, self.__class__):
      return self.AsBytes() < other.AsBytes()
    else:
      return self.AsBytes() < other

  def __eq__(self, other):
    if isinstance(other, self.__class__):
      return self.AsBytes() == other.AsBytes()
    else:
      return self.AsBytes() == other

  def __len__(self):
    return len(self.AsBytes())


class RDFZippedBytes(RDFBytes):
  """Zipped bytes sequence."""

  def Uncompress(self):
    if self:
      return zlib.decompress(self._value)
    else:
      return b""


@functools.total_ordering
class RDFString(RDFPrimitive):
  """Represent a simple string."""

  protobuf_type = "string"

  # TODO(hanuszczak): Allow initializing from arbitrary `unicode`-able object.
  def __init__(self, initializer=None):
    if initializer is None:
      super().__init__("")
    if isinstance(initializer, RDFString):
      super().__init__(str(initializer))
    elif isinstance(initializer, bytes):
      super().__init__(initializer.decode("utf-8"))
    elif isinstance(initializer, str):
      super().__init__(initializer)
    elif initializer is not None:
      message = "Unexpected initializer `%s` of type `%s`"
      message %= (initializer, type(initializer))
      raise TypeError(message)

  def format(self, *args, **kwargs):  # pylint: disable=invalid-name
    return self._value.format(*args, **kwargs)

  def split(self, *args, **kwargs):  # pylint: disable=invalid-name
    return self._value.split(*args, **kwargs)

  def __str__(self) -> str:
    return self._value

  def __hash__(self):
    return hash(self._value)

  def __getitem__(self, item):
    return self._value.__getitem__(item)

  def __len__(self):
    return len(self._value)

  def __eq__(self, other):
    if isinstance(other, RDFString):
      return self._value == other._value  # pylint: disable=protected-access

    if isinstance(other, str):
      return self._value == other

    # TODO(hanuszczak): Comparing `RDFString` and `bytes` should result in type
    # error. For now we allow it because too many tests still use non-unicode
    # string literals.
    if isinstance(other, bytes):
      return self._value.encode("utf-8") == other

    return NotImplemented

  def __lt__(self, other):
    if isinstance(other, RDFString):
      return self._value < other._value  # pylint: disable=protected-access

    if isinstance(other, str):
      return self._value < other

    # TODO(hanuszczak): Comparing `RDFString` and `bytes` should result in type
    # error. For now we allow it because too many tests still use non-unicode
    # string literals.
    if isinstance(other, bytes):
      return self._value.encode("utf-8") < other

    return NotImplemented

  @classmethod
  def FromSerializedBytes(cls, value: bytes):
    precondition.AssertType(value, bytes)
    return cls(value)

  @classmethod
  def FromWireFormat(cls, value):
    return cls.FromHumanReadable(value)

  @classmethod
  def FromHumanReadable(cls, string: str):
    precondition.AssertType(string, str)
    return cls(string)

  def SerializeToBytes(self):
    return self._value.encode("utf-8")

  def SerializeToWireFormat(self):
    return self._value


# TODO(hanuszczak): This class should provide custom method for parsing from
# human readable strings (and arguably should not derive from `RDFBytes` at
# all).
class HashDigest(RDFBytes):
  """Binary hash digest with hex string representation."""

  protobuf_type = "bytes"

  def HexDigest(self) -> str:
    return text.Hexify(self._value)

  def __str__(self) -> str:
    return self.HexDigest()

  def __hash__(self):
    return hash(self._value)

  # TODO(hanuszczak): This is a terrible equality definition.
  def __eq__(self, other):
    if isinstance(other, HashDigest):
      return self._value == other._value  # pylint: disable=protected-access
    if isinstance(other, bytes):
      return self._value == other
    if isinstance(other, str):
      return str(self) == other
    return NotImplemented

  def __ne__(self, other):
    return not self == other


@functools.total_ordering
class RDFInteger(RDFPrimitive):
  """Represent an integer."""

  protobuf_type = "integer"

  @staticmethod
  def IsNumeric(value):
    return isinstance(value, (int, float, RDFInteger))

  def __init__(self, initializer=None):
    if initializer is None:
      super().__init__(0)
    else:
      super().__init__(int(initializer))

  def SerializeToBytes(self) -> bytes:
    return str(self._value).encode("ascii")

  @classmethod
  def FromSerializedBytes(cls, value: bytes):
    precondition.AssertType(value, bytes)

    if value:
      return cls(int(value))
    else:
      return cls(0)

  @classmethod
  def FromHumanReadable(cls, string: str):
    precondition.AssertType(string, str)
    return cls(int(string))

  def __str__(self) -> str:
    return str(self._value)

  @classmethod
  def FromWireFormat(cls, value):
    return cls(initializer=value)

  def SerializeToWireFormat(self):
    """Use varint to store the integer."""
    return self._value

  def __long__(self):
    return self._value

  def __int__(self):
    return self._value

  def __float__(self):
    return float(self._value)

  def __index__(self):
    return self._value

  def __lt__(self, other):
    return self._value < other

  def __and__(self, other):
    return self._value & other

  def __rand__(self, other):
    return self._value & other

  def __or__(self, other):
    return self._value | other

  def __ror__(self, other):
    return self._value | other

  def __add__(self, other):
    return self._value + other

  def __radd__(self, other):
    return self._value + other

  def __sub__(self, other):
    return self._value - other

  def __rsub__(self, other):
    return other - self._value

  def __mul__(self, other):
    return self._value * other

  def __rmul__(self, other):
    return self._value * other

  def __div__(self, other):
    return self._value.__div__(other)

  def __truediv__(self, other):
    return self._value.__truediv__(other)

  def __floordiv__(self, other):
    return self._value.__floordiv__(other)

  def __hash__(self):
    return hash(self._value)


@functools.total_ordering
class RDFDatetime(RDFPrimitive):
  """A date and time internally stored in MICROSECONDS."""

  converter = 1000000
  protobuf_type = "unsigned_integer"

  def __init__(self, initializer=None):
    if initializer is None:
      super().__init__(0)
    elif isinstance(initializer, RDFDatetime):
      super().__init__(
          initializer.AsMicrosecondsSinceEpoch() * self.converter // 1000000
      )
    elif isinstance(initializer, float):
      raise TypeError("Float initialization not supported.")
    elif isinstance(initializer, (RDFInteger, int)):
      super().__init__(int(initializer))
    else:
      raise InitializeError(
          "Unknown initializer for RDFDateTime: %s %s."
          % (type(initializer), initializer)
      )

  @classmethod
  def FromWireFormat(cls, value: int):
    return cls(initializer=value)

  def SerializeToWireFormat(self) -> int:
    """Use varint to store the integer."""
    return self._value

  def SerializeToBytes(self) -> bytes:
    return str(self._value).encode("ascii")

  @classmethod
  def FromSerializedBytes(cls, value: bytes):
    precondition.AssertType(value, bytes)

    if value:
      return cls(int(value))
    else:
      return cls(0)

  @classmethod
  def Now(cls):
    return cls(int(time.time() * cls.converter))

  def Format(self, fmt: str) -> str:
    """Return the value as a string formatted as per strftime semantics."""
    precondition.AssertType(fmt, str)

    stime = time.gmtime(self._value / self.converter)
    return time.strftime(fmt, stime)

  def __str__(self) -> str:
    """Return the date in human readable (UTC)."""
    # TODO: Display microseconds if applicable.
    return self.Format("%Y-%m-%d %H:%M:%S")

  def AsDatetime(self) -> datetime.datetime:
    """Returns the datetime as Python `datetime` without any timezone set."""
    return datetime.datetime.utcfromtimestamp(self._value / self.converter)

  def AsDatetimeUTC(self) -> datetime.datetime:
    """Returns the datetime as a Python `datetime` with UTC timezone."""
    return self.AsDatetime().replace(tzinfo=datetime.timezone.utc)

  def AsProtoTimestamp(self) -> timestamp_pb2.Timestamp:
    """Returns the datetime as standard Protocol Buffers timestamp message."""
    micros = self.AsMicrosecondsSinceEpoch()

    timestamp = timestamp_pb2.Timestamp()
    timestamp.seconds = micros // 1_000_000
    timestamp.nanos = (micros % 1_000_000) * 1_000
    return timestamp

  def AsSecondsSinceEpoch(self):
    return self._value // self.converter

  def AsMicrosecondsSinceEpoch(self):
    return self._value * 1000000 // self.converter

  @classmethod
  def FromSecondsSinceEpoch(cls, value):
    # Convert to int in case we get fractional seconds with higher
    # resolution than what this class supports.
    return cls(int(value * cls.converter))

  @classmethod
  def FromMicrosecondsSinceEpoch(cls, value):
    precondition.AssertType(value, int)
    return cls(value * cls.converter // 1000000)

  @classmethod
  def FromDatetime(cls, value):
    seconds = calendar.timegm(value.utctimetuple())
    return cls(
        (seconds * cls.converter)
        + (value.microsecond * cls.converter // 1000000)
    )

  @classmethod
  def FromProtoTimestamp(
      cls,
      timestamp: timestamp_pb2.Timestamp,
  ) -> "RDFDatetime":
    """Converts Protocol Buffers `Timestamp` instances to datetime objects.

    Args:
      timestamp: A Protocol Buffers `Timestamp` instance to convert.

    Returns:
      A corresponding RDF datetime object.
    """
    micros = timestamp.seconds * 1_000_000 + timestamp.nanos // 1_000
    return RDFDatetime.FromMicrosecondsSinceEpoch(micros)

  @classmethod
  def FromDate(cls, value):
    seconds = calendar.timegm(value.timetuple())
    return cls(seconds * cls.converter)

  @classmethod
  def Lerp(cls, t, start_time, end_time):
    """Interpolates linearly between two datetime values.

    Args:
      t: An interpolation "progress" value.
      start_time: A value for t = 0.
      end_time: A value for t = 1.

    Returns:
      An interpolated `RDFDatetime` instance.

    Raises:
      TypeError: If given time values are not instances of `RDFDatetime`.
      ValueError: If `t` parameter is not between 0 and 1.
    """
    if not (
        isinstance(start_time, RDFDatetime)
        and isinstance(end_time, RDFDatetime)
    ):
      raise TypeError("Interpolation of non-datetime values")

    if not 0.0 <= t <= 1.0:
      raise ValueError("Interpolation progress does not belong to [0.0, 1.0]")

    return cls(round((1 - t) * start_time._value + t * end_time._value))  # pylint: disable=protected-access

  def __add__(self, other):
    if isinstance(other, float):
      raise TypeError("Float initialization not supported.")
    elif isinstance(other, int):
      # Assume other is in seconds
      other_microseconds = int(other * self.converter)
      return self.__class__(self._value + other_microseconds)
    elif isinstance(other, (DurationSeconds, Duration)):
      self_us = self.AsMicrosecondsSinceEpoch()
      duration_us = other.microseconds
      return self.__class__.FromMicrosecondsSinceEpoch(self_us + duration_us)

    return NotImplemented

  def __sub__(self, other):
    if isinstance(other, float):
      raise TypeError("Float initialization not supported.")
    elif isinstance(other, int):
      # Assume other is in seconds
      other_microseconds = int(other * self.converter)
      return self.__class__(self._value - other_microseconds)
    elif isinstance(other, (DurationSeconds, Duration)):
      self_us = self.AsMicrosecondsSinceEpoch()
      duration_us = other.microseconds
      return self.__class__.FromMicrosecondsSinceEpoch(self_us - duration_us)
    elif isinstance(other, RDFDatetime):
      diff_us = (
          self.AsMicrosecondsSinceEpoch() - other.AsMicrosecondsSinceEpoch()
      )
      return Duration.From(diff_us, MICROSECONDS)

    return NotImplemented

  @classmethod
  def FromHumanReadable(cls, string: str, eoy=False):
    """Parses a human readable string of a timestamp (in local time).

    Args:
      string: The string to parse.
      eoy: If True, sets the default value to the end of the year. Usually this
        method returns a timestamp where each field that is not present in the
        given string is filled with values from the date January 1st of the
        current year, midnight. Sometimes it makes more sense to compare against
        the end of a period so if eoy is set, the default values are copied from
        the 31st of December of the current year, 23:59h.

    Returns:
      A new instance based on the given string.
    """
    # TODO(hanuszczak): This method should accept only unicode literals.
    # TODO(hanuszczak): Date can come either as a single integer (which we
    # interpret as a timestamp) or as a really human readable thing such as
    # '2000-01-01 13:37'. This is less than ideal (since timestamps are not
    # really "human readable") and should be fixed in the future.
    try:
      return cls(int(string))
    except ValueError:
      pass

    # By default assume the time is given in UTC.
    # pylint: disable=g-tzinfo-datetime
    if eoy:
      default = datetime.datetime(
          time.gmtime().tm_year, 12, 31, 23, 59, tzinfo=dateutil.tz.tzutc()
      )
    else:
      default = datetime.datetime(
          time.gmtime().tm_year, 1, 1, 0, 0, tzinfo=dateutil.tz.tzutc()
      )
    # pylint: enable=g-tzinfo-datetime

    timestamp = parser.parse(string, default=default)

    raw = calendar.timegm(timestamp.utctimetuple()) * cls.converter
    return cls(raw)

  def Floor(self, interval):
    precondition.AssertType(interval, Duration)
    seconds = (
        self.AsSecondsSinceEpoch()
        // interval.ToInt(SECONDS)
        * interval.ToInt(SECONDS)
    )
    return self.FromSecondsSinceEpoch(seconds)

  def __hash__(self):
    return hash(self._value)

  def __eq__(self, other):
    if isinstance(other, RDFDatetime):
      return self.AsMicrosecondsSinceEpoch() == other.AsMicrosecondsSinceEpoch()
    else:
      return self._value == other

  def __lt__(self, other):
    if isinstance(other, RDFDatetime):
      return self.AsMicrosecondsSinceEpoch() < other.AsMicrosecondsSinceEpoch()
    else:
      return self._value < other

  def __int__(self):
    return self._value

  def AbsDiff(self, other: "RDFDatetime") -> "Duration":
    if self > other:
      return self - other
    return other - self


class RDFDatetimeSeconds(RDFDatetime):
  """A DateTime class which is stored in whole seconds."""

  converter = 1


# Constants used as time unit in Duration methods.
MICROSECONDS = 1
MILLISECONDS = 1000
SECONDS = 1000 * MILLISECONDS
MINUTES = 60 * SECONDS
HOURS = 60 * MINUTES
DAYS = 24 * HOURS
WEEKS = 7 * DAYS

_DURATION_RE = re.compile(r"(?P<number>\d+) ?(?P<unit>[a-z]{1,2})?")


@functools.total_ordering
class Duration(RDFPrimitive):
  """Absolute duration between instants in time with microsecond precision.

  The duration is stored as non-negative integer, guaranteeing microsecond
  precision up to MAX_UINT64 microseconds (584k years).
  """

  protobuf_type = "unsigned_integer"

  _DIVIDERS = dict((
      ("w", WEEKS),
      ("d", DAYS),
      ("h", HOURS),
      ("m", MINUTES),
      ("s", SECONDS),
      ("ms", MILLISECONDS),
      ("us", MICROSECONDS),
  ))

  def __init__(self, initializer=None):
    """Instantiates a new microsecond-based Duration.

    Args:
      initializer: Integer specifying microseconds, or another Duration to copy.
        If None, Duration will be set to 0. Given a negative integer, its
        absolute (positive) value will be stored.
    """
    if isinstance(initializer, Duration):
      if initializer.microseconds < 0:
        raise ValueError("Negative Duration (%s ms)" % initializer.microseconds)
      super().__init__(initializer.microseconds)
    elif isinstance(initializer, (int, RDFInteger)):
      if int(initializer) < 0:
        raise ValueError("Negative Duration (%s s)" % initializer)
      super().__init__(int(initializer))
    elif isinstance(initializer, str):
      super().__init__(self._ParseText(initializer, default_unit=None))
    elif initializer is None:
      super().__init__(0)
    else:
      message = "Unsupported initializer `{value}` of type `{type}`"
      raise TypeError(message.format(value=initializer, type=type(initializer)))

  @classmethod
  def From(cls, value: Union[int, float], timeunit: int):
    """Returns a new Duration given a timeunit and value.

    Args:
      value: A number specifying the value of the duration.
      timeunit: A unit of time ranging from rdfvalue.MICROSECONDS to
        rdfvalue.WEEKS.

    Examples:
      >>> Duration.From(50, MICROSECONDS)
      <Duration 50 us>

      >>> Duration.From(120, SECONDS)
      <Duration 2 m>

    Returns:
      A new Duration.
    """
    return cls(int(timeunit * value))

  @classmethod
  def FromWireFormat(cls, value):
    precondition.AssertType(value, int)
    return cls(value)

  def SerializeToWireFormat(self):
    """See base class."""
    return self.microseconds

  @classmethod
  def FromSerializedBytes(cls, value: bytes):
    precondition.AssertType(value, bytes)

    if not value:
      return cls(0)

    try:
      raw = abs(int(value))
    except ValueError as e:
      raise DecodeError(e) from e

    return cls(raw)

  def SerializeToBytes(self) -> bytes:
    """See base class."""
    # Technically, equal to ascii encoding, since str(self._value) only contains
    # the digits 0-9.
    return str(self.microseconds).encode("utf-8")

  def __repr__(self):
    return "<{} {}>".format(type(self).__name__, self)

  def __str__(self) -> str:
    if self._value == 0:
      return "0 us"
    for label, divider in self._DIVIDERS.items():
      if self._value % divider == 0:
        return "%d %s" % (self._value // divider, label)
    return "%d us" % self._value  # Make pytype happy.

  def __add__(self, other):
    if isinstance(other, Duration):
      return self.__class__.From(
          self.microseconds + other.microseconds, MICROSECONDS
      )
    else:
      return NotImplemented

  def __sub__(self, other):
    if isinstance(other, Duration):
      return self.__class__.From(
          self.microseconds - other.microseconds, MICROSECONDS
      )
    else:
      return NotImplemented

  def __mul__(self, other):
    if isinstance(other, int):
      return self.__class__.From(self.microseconds * other, MICROSECONDS)
    else:
      return NotImplemented

  def __rmul__(self, other):
    return self.__mul__(other)

  def __lt__(self, other):
    if isinstance(other, Duration):
      return self.microseconds < other.microseconds
    else:
      return NotImplemented

  def __eq__(self, other):
    if isinstance(other, Duration):
      return self.microseconds == other.microseconds
    else:
      return NotImplemented

  def __abs__(self):
    return self

  def ToInt(self, timeunit: int) -> int:
    """Returns the duration as truncated integer, converted to the time unit.

    All fractions are truncated. To preserve them, use `toFractional()`.

    Examples:
      >>> Duration.From(2, WEEKS).ToInt(DAYS)
      14

      >>> Duration.From(100, SECONDS).ToInt(SECONDS)
      100

      >>> Duration.From(6, DAYS).ToInt(WEEKS)
      0

    Args:
      timeunit: A unit of time ranging from rdfvalue.MICROSECONDS to
        rdfvalue.WEEKS.

    Returns:
      An integer, representing the duration in the specific unit, truncating
      fractions.
    """
    return self.microseconds // timeunit

  def ToFractional(self, timeunit: int) -> float:
    """Returns the duration as float, converted to the given time unit.

    Examples:
      >>> Duration.From(30, SECONDS).ToFractional(MINUTES)
      0.5

      >>> Duration.From(100, SECONDS).ToFractional(SECONDS)
      100.0

      >>> Duration.From(6, MINUTES).ToFractional(HOURS)
      0.1

    Args:
      timeunit: A unit of time ranging from rdfvalue.MICROSECONDS to
        rdfvalue.WEEKS.

    Returns:
      A float, representing the duration in the specific unit, including
      fractions.
    """
    return self.microseconds / timeunit

  def AsTimedelta(self) -> datetime.timedelta:
    """Returns a standard `timedelta` object corresponding to the duration."""
    return datetime.timedelta(microseconds=self.ToInt(MICROSECONDS))

  @property
  def microseconds(self):
    return self._value

  @classmethod
  def FromHumanReadable(cls, string: str):
    """See base class."""
    return cls(cls._ParseText(string, default_unit=None))

  @classmethod
  def _ParseText(cls, string: str, default_unit: Optional[str]) -> int:
    """Parses a textual representation of a duration."""
    precondition.AssertType(string, str)

    if not string:
      return 0

    matches = _DURATION_RE.match(string)
    if matches is None:
      raise ValueError("Could not parse duration {!r}.".format(string))

    number = int(matches.group("number"))
    unit_string = matches.group("unit")

    if unit_string is None:
      unit_string = default_unit

    try:
      unit_multiplier = cls._DIVIDERS[unit_string]
    except KeyError as ex:
      raise ValueError(
          "Invalid unit {!r} for duration in {!r}. Expected any of {}.".format(
              unit_string, string, ", ".join(cls._DIVIDERS)
          )
      ) from ex

    return number * unit_multiplier

  def Expiry(self, base_time=None):
    if base_time is None:
      base_time = RDFDatetime.Now()
    return base_time + self


class DurationSeconds(Duration):
  """Duration that is (de)serialized with second-precision.

  This class exists for compatibility purposes and to keep certain API fields
  simple. For most uses, please prefer `Duration` directly.
  """

  def __init__(self, initializer: Any = None):
    if isinstance(initializer, (int, RDFInteger)):
      initializer = int(initializer) * SECONDS
    elif isinstance(initializer, str):
      initializer = self._ParseText(initializer, default_unit="s")
    super().__init__(initializer)

  def SerializeToBytes(self) -> bytes:
    """See base class."""
    return str(self.ToInt(SECONDS)).encode("utf-8")

  def SerializeToWireFormat(self):
    """See base class."""
    return self.ToInt(SECONDS)

  @classmethod
  def FromHumanReadable(cls, string: str):
    precondition.AssertType(string, str)
    return cls(string)

  def __str__(self) -> str:
    if self.microseconds == 0:
      return "0 s"
    return super().__str__()

  @classmethod
  def From(cls, value: Union[int, float], timeunit: int):
    """See base class."""
    return cls(int(timeunit * value // SECONDS))


class ByteSize(RDFInteger):
  """A size for bytes allowing standard unit prefixes.

  We use the standard IEC 60027-2 A.2 and ISO/IEC 80000:
  Binary units (powers of 2): Ki, Mi, Gi
  SI units (powers of 10): k, m, g
  """

  protobuf_type = "unsigned_integer"

  DIVIDERS = dict((
      ("", 1),
      ("k", 1000),
      ("m", 1000**2),
      ("g", 1000**3),
      ("ki", 1024),
      ("mi", 1024**2),
      ("gi", 1024**3),
  ))

  REGEX = re.compile("^([0-9.]+) ?([kmgi]*)b?$", re.I)

  def __init__(self, initializer=None):
    if isinstance(initializer, ByteSize):
      super().__init__(initializer._value)  # pylint: disable=protected-access
    elif isinstance(initializer, str):
      super().__init__(self._ParseText(initializer))
    elif isinstance(initializer, (int, float)):
      super().__init__(initializer)
    elif isinstance(initializer, RDFInteger):
      super().__init__(int(initializer))
    elif initializer is None:
      super().__init__(0)
    else:
      raise InitializeError(
          "Unknown initializer for ByteSize: %s." % type(initializer)
      )

  def __str__(self):
    if self._value >= 1024**3:
      unit = "GiB"
      value = self._value / 1024**3
    elif self._value >= 1024**2:
      unit = "MiB"
      value = self._value / 1024**2
    elif self._value >= 1024:
      unit = "KiB"
      value = self._value / 1024
    else:
      return "{} B".format(self._value)

    return "{value:.1f} {unit}".format(value=value, unit=unit)

  @classmethod
  def FromHumanReadable(cls, string: str):
    return cls(cls._ParseText(string))

  @classmethod
  def _ParseText(cls, string: str) -> int:
    """Parses a textual representation of the number of bytes."""
    if not string:
      return 0

    match = cls.REGEX.match(string.strip().lower())
    if not match:
      raise DecodeError("Unknown specification for ByteSize %s" % string)

    multiplier = cls.DIVIDERS.get(match.group(2))
    if not multiplier:
      raise DecodeError("Invalid multiplier %s" % match.group(2))

    # The value may be represented as a float, but if it's not, don't lose
    # accuracy.
    value = match.group(1)
    if "." in value:
      value = float(value)
    else:
      value = int(value)

    return int(value * multiplier)


@functools.total_ordering
class RDFURN(RDFPrimitive):
  """An object to abstract URL manipulation."""

  protobuf_type = "string"

  # Careful when changing this value, this is hardcoded a few times in this
  # class for performance reasons.
  scheme = "aff4"

  def __init__(self, initializer=None):
    """Constructor.

    Args:
      initializer: A string or another RDFURN.
    """
    # This is a shortcut that is a bit faster than the standard way of
    # using the RDFValue constructor to make a copy of the class. For
    # RDFURNs that way is a bit slow since it would try to normalize
    # the path again which is not needed - it comes from another
    # RDFURN so it is already in the correct format.

    if initializer is None:
      super().__init__("")
      return

    if isinstance(initializer, RDFURN):
      # Make a direct copy of the other object
      super().__init__(cast(RDFURN, initializer).Path())
      return

    precondition.AssertType(initializer, (bytes, str))

    if isinstance(initializer, bytes):
      initializer = initializer.decode("utf-8")

    super().__init__(self._Normalize(initializer))

  @classmethod
  def _Normalize(cls, string):
    if string.startswith("aff4:/"):
      string = string[5:]
    return utils.NormalizePath(string)

  @classmethod
  def FromSerializedBytes(cls, value: bytes):
    precondition.AssertType(value, bytes)
    return cls(value)

  @classmethod
  def FromWireFormat(cls, value):
    # TODO(hanuszczak): We should just assign the `self._value` here
    # instead of including all of the parsing magic since the data store values
    # should be normalized already. But sadly this is not the case and for now
    # we have to deal with unnormalized values as well.
    return cls(value)

  @classmethod
  def FromHumanReadable(cls, string: str):
    precondition.AssertType(string, str)
    return cls(string)

  def SerializeToBytes(self) -> bytes:
    return str(self).encode("utf-8")

  def SerializeToWireFormat(self) -> str:
    return str(self)

  def Dirname(self):
    return posixpath.dirname(self._value)

  def Basename(self):
    return posixpath.basename(self.Path())

  def Add(self, path):
    """Add a relative stem to the current value and return a new RDFURN.

    If urn is a fully qualified URN, replace the current value with it.

    Args:
      path: A string containing a relative path.

    Returns:
       A new RDFURN that can be chained.

    Raises:
       ValueError: if the path component is not a string.
    """
    if not isinstance(path, str):
      raise ValueError(
          "Only strings should be added to a URN, not %s" % path.__class__
      )
    return self.__class__(utils.JoinPath(self._value, path))

  def __str__(self) -> str:
    return "aff4:%s" % self._value

  # Required, because in Python 3 overriding `__eq__` nullifies `__hash__`.
  __hash__ = RDFPrimitive.__hash__

  def __eq__(self, other):
    if isinstance(other, str):
      other = self.__class__(other)

    elif other is None:
      return False

    elif not isinstance(other, RDFURN):
      return NotImplemented

    return self._value == other.Path()

  def __bool__(self):
    return bool(self._value)

  def __lt__(self, other):
    return self._value < other

  def Path(self):
    """Return the path of the urn."""
    return self._value

  def Split(self, count=None):
    """Returns all the path components.

    Args:
      count: If count is specified, the output will be exactly this many path
        components, possibly extended with the empty string. This is useful for
        tuple assignments without worrying about ValueErrors:  namespace, path =
        urn.Split(2)

    Returns:
      A list of path components of this URN.
    """
    if count:
      result = list(filter(None, self._value.split("/", count)))
      while len(result) < count:
        result.append("")

      return result

    else:
      return list(filter(None, self._value.split("/")))

  def RelativeName(self, volume):
    """Given a volume URN return the relative URN as a unicode string.

    We remove the volume prefix from our own.
    Args:
      volume: An RDFURN or fully qualified url string.

    Returns:
      A string of the url relative from the volume or None if our URN does not
      start with the volume prefix.
    """
    string_url = utils.SmartUnicode(self)
    volume_url = utils.SmartUnicode(volume)
    if string_url.startswith(volume_url):
      result = string_url[len(volume_url) :]
      # This must always return a relative path so we strip leading "/"s. The
      # result is always a unicode string.
      return result.lstrip("/")

    return None

  def __repr__(self):
    return "<%s>" % self


class Subject(RDFURN):
  """A pseudo attribute representing the subject of an AFF4 object."""


DEFAULT_FLOW_QUEUE = RDFURN("F")


class SessionID(RDFURN):
  """An rdfvalue object that represents a session_id."""

  def __init__(
      self,
      initializer=None,
      base="aff4:/flows",
      queue=DEFAULT_FLOW_QUEUE,
      flow_name=None,
  ):
    """Constructor.

    Args:
      initializer: A string or another RDFURN.
      base: The base namespace this session id lives in.
      queue: The queue to use.
      flow_name: The name of this flow or its random id.

    Raises:
      InitializeError: The given URN cannot be converted to a SessionID.
    """
    if initializer is None:
      # This SessionID is being constructed from scratch.
      if flow_name is None:
        flow_name = random.UInt32()

      if isinstance(flow_name, int):
        initializer = RDFURN(base).Add("%s:%X" % (queue.Basename(), flow_name))
      else:
        initializer = RDFURN(base).Add("%s:%s" % (queue.Basename(), flow_name))
    else:
      if isinstance(initializer, RDFURN):
        try:
          self.ValidateID(initializer.Basename())
        except ValueError as e:
          raise InitializeError(
              "Invalid URN for SessionID: %s, %s" % (initializer, e)
          ) from e

    super().__init__(initializer=initializer)

  def Add(self, path):
    # Adding to a SessionID results in a normal RDFURN.
    return RDFURN(self).Add(path)

  @classmethod
  def ValidateID(cls, id_str):
    # This check is weaker than it could be because we allow queues called
    # "DEBUG-user1" and IDs like "TransferStore". We also have to allow
    # flows session ids like H:123456:hunt.
    allowed_re = re.compile(r"^[-0-9a-zA-Z]+(:[0-9a-zA-Z]+){0,2}$")
    if not allowed_re.match(id_str):
      raise ValueError("Invalid SessionID: %s" % id_str)


# TODO(hanuszczak): Remove this class.
class FlowSessionID(SessionID):
  pass



#!/usr/bin/env python
"""A module with utilities for a very simple serialization format."""

from collections.abc import Iterator
import gzip
import io
import os
import struct

from grr_response_core.lib.util import chunked

DEFAULT_CHUNK_SIZE = 1 * 1024 * 1024  # 1 MiB.


def Serialize(
    stream: Iterator[bytes],
    chunk_size: int = DEFAULT_CHUNK_SIZE,
) -> Iterator[bytes]:
  """Serializes a stream of data to the stream of chunks.

  Args:
    stream: A stream of data to serialize.
    chunk_size: An (optional) approximate size of a chunk in bytes. Every non-
      final chunk will be slightly bigger than the specified number, but this
      should be negligible.

  Yields:
    Serialized chunks (in the gzchunked format).
  """
  while True:
    buf = io.BytesIO()
    buf_entry_count = 0

    with gzip.GzipFile(fileobj=buf, mode="wb") as filedesc:
      for data in stream:
        chunked.Write(filedesc, data)  # pytype: disable=wrong-arg-types
        buf_entry_count += 1

        if buf.tell() >= chunk_size:
          break

    if buf_entry_count == 0:
      break

    yield buf.getvalue()


def Deserialize(stream: Iterator[bytes]) -> Iterator[bytes]:
  """Deserializes a stream a chunks into a stream of data.

  Args:
    stream: A stream of serialized chunks (in the gzchunked format).

  Yields:
    A stream of deserialized data.
  """
  for chunk in stream:
    buf = io.BytesIO(chunk)

    with gzip.GzipFile(fileobj=buf, mode="rb") as filedesc:
      filedesc.seek(0, os.SEEK_END)
      fd_size = filedesc.tell()
      filedesc.seek(0, os.SEEK_SET)

      while True:
        data = chunked.Read(filedesc, max_chunk_size=fd_size)  # pytype: disable=wrong-arg-types
        if data is None:
          break

        fd_size -= len(data)
        yield data


_UINT64 = struct.Struct("!Q")  # Network-endian 64-bit unsigned integer format.



#!/usr/bin/env python
"""Provides conversion functions to be used during RDFProtoStruct migration."""

from grr_response_core.lib.rdfvalues import large_file as rdf_large_file
from grr_response_proto import large_file_pb2


def ToProtoCollectLargeFileArgs(
    rdf: rdf_large_file.CollectLargeFileArgs,
) -> large_file_pb2.CollectLargeFileArgs:
  return rdf.AsPrimitiveProto()


def ToRDFCollectLargeFileArgs(
    proto: large_file_pb2.CollectLargeFileArgs,
) -> rdf_large_file.CollectLargeFileArgs:
  return rdf_large_file.CollectLargeFileArgs.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoCollectLargeFileResult(
    rdf: rdf_large_file.CollectLargeFileResult,
) -> large_file_pb2.CollectLargeFileResult:
  return rdf.AsPrimitiveProto()


def ToRDFCollectLargeFileResult(
    proto: large_file_pb2.CollectLargeFileResult,
) -> rdf_large_file.CollectLargeFileResult:
  return rdf_large_file.CollectLargeFileResult.FromSerializedBytes(
      proto.SerializeToString()
  )



#!/usr/bin/env python
"""A module with utilities for working with I/O."""

from collections.abc import Iterator
import io
from typing import IO


def Chunk(stream: IO[bytes], size: int) -> Iterator[bytes]:
  """Divides given stream into chunks of specified size.

  Args:
    stream: A file-like object to chunk.
    size: Size of individual chunks (the last one might be smaller).

  Yields:
    Chunks with content of the stream.
  """
  if size <= 0:
    raise ValueError(f"Non-positive chunk size: {size}")

  while True:
    data = stream.read(size)
    if not data:
      break

    yield data


def Unchunk(chunks: Iterator[bytes]) -> IO[bytes]:
  """Joins chunks of a file to a file-like object.

  Args:
    chunks: An iterator yielding chunks of the file.

  Returns:
    A file-like object that
  """
  # For some reason the linter doesn't understand that `RawIOBase` implements
  # the `IO[bytes]` interface and complains.
  return io.BufferedReader(_Unchunked(chunks))  # pylint: disable=abstract-class-instantiated


class _Unchunked(io.RawIOBase, IO[bytes]):  # pytype: disable=signature-mismatch  # overriding-return-type-checks
  """A raw file-like object that reads chunk stream on demand."""

  def __init__(self, chunks: Iterator[bytes]) -> None:
    """Initializes the object."""
    super().__init__()
    self._chunks = chunks
    self._buf = io.BytesIO()

  def readable(self) -> bool:
    return True

  def readall(self) -> bytes:
    return b"".join(self._chunks)

  def readinto(self, buf: bytearray) -> int:
    if self._buf.tell() == len(self._buf.getbuffer()):
      self._buf.seek(0, io.SEEK_SET)
      self._buf.truncate()
      self._buf.write(next(self._chunks, b""))
      self._buf.seek(0, io.SEEK_SET)

    return self._buf.readinto(buf)



#!/usr/bin/env python
"""Library for processing of artifacts.

This file contains non-GRR specific pieces of artifact processing and is
intended to end up as an independent library.
"""

from collections.abc import Sequence
import re

from grr_response_proto import knowledge_base_pb2


class Error(Exception):
  """Base exception."""


class ConditionError(Error):
  """An invalid artifact condition was specified."""


class ArtifactProcessingError(Error):
  """Unable to process artifact."""


class KnowledgeBaseUninitializedError(Error):
  """Attempt to process artifact without a valid Knowledge Base."""


class KnowledgeBaseAttributesMissingError(Error):
  """Knowledge Base is missing key attributes."""


INTERPOLATED_REGEX = re.compile(r"%%([^%]+?)%%")


def ExpandKnowledgebaseWindowsEnvVars(
    unexpanded_kb: knowledge_base_pb2.KnowledgeBase,
) -> knowledge_base_pb2.KnowledgeBase:
  """Expands all Windows environment variable values in the given knowledgebase.

  Unexpanded values can contain references to other environment variables, e.g.
  `%SystemRoot/System32`. Such references are expanded using knowledgebase
  values recursively, e.g. the above could be expanded to `C:/System32`.

  If an environment variable value contains a reference that cannot be expanded,
  this function will not raise but rather leave it in unexpanded form (similarly
  to what Windows shell does).

  If unexpanded references form a cycle, this function will raise.

  Args:
    unexpanded_kb: A knowledgebase with environment variables to expand.

  Returns:
    A knowledgebase in which all environment variables are expanded.
  """
  if unexpanded_kb.os != "Windows":
    raise ValueError(f"Invalid system: {unexpanded_kb.os!r}")

  kb = knowledge_base_pb2.KnowledgeBase(
      environ_path="%SystemRoot%\\;%SystemRoot%\\System32\\;%SystemRoot%\\System32\\wbem\\",
      environ_temp="%SystemRoot%\\TEMP",
      environ_allusersappdata="%ProgramData%",
      environ_allusersprofile="%ProgramData%",
      environ_commonprogramfiles="%ProgramFiles%\\Common Files",
      environ_commonprogramfilesx86="%ProgramFiles(x86)%\\Common Files",
      environ_comspec="%SystemRoot%\\System32\\cmd.exe",
      environ_driverdata="%SystemRoot%\\System32\\Drivers\\DriverData",
      environ_programfiles="%SystemDrive%\\Program Files",
      environ_programfilesx86="%SystemDrive%\\Program Files (x86)",
      environ_programdata="%SystemDrive%\\ProgramData",
      environ_systemdrive="C:",
      environ_systemroot="%SystemDrive%\\Windows",
      environ_windir="%SystemRoot%",
  )
  kb.MergeFrom(unexpanded_kb)

  already_expanded_env_var_refs: dict[str, str] = dict()
  currently_expanded_env_var_refs: set[str] = set()

  def Expand(unexpanded: str) -> str:
    expanded = ""
    offset = 0

    for match in re.finditer("%[^%]+?%", unexpanded):
      env_var_ref = match.group(0).upper()

      expanded += unexpanded[offset : match.start()]
      offset += match.end()

      if env_var_ref in already_expanded_env_var_refs:
        expanded += already_expanded_env_var_refs[env_var_ref]
        continue

      if env_var_ref in currently_expanded_env_var_refs:
        raise ValueError(f"Circular dependency involving {env_var_ref!r}")

      if env_var_ref == "%PATH%":
        value = kb.environ_path
      elif env_var_ref == "%TEMP%":
        value = kb.environ_temp
      elif env_var_ref == "%ALLUSERSAPPDATA%":
        value = kb.environ_allusersappdata
      elif env_var_ref == "%ALLUSERSPROFILE%":
        value = kb.environ_allusersprofile
      elif env_var_ref == "%COMMONPROGRAMFILES%":
        value = kb.environ_commonprogramfiles
      elif env_var_ref == "%COMMONPROGRAMFILES(X86)%":
        value = kb.environ_commonprogramfilesx86
      elif env_var_ref == "%COMSPEC%":
        value = kb.environ_comspec
      elif env_var_ref == "%DRIVERDATA%":
        value = kb.environ_driverdata
      elif env_var_ref == "%PROGRAMFILES%":
        value = kb.environ_programfiles
      elif env_var_ref == "%PROGRAMFILES(X86)%":
        value = kb.environ_programfilesx86
      elif env_var_ref == "%PROGRAMDATA%":
        value = kb.environ_programdata
      elif env_var_ref == "%SYSTEMDRIVE%":
        value = kb.environ_systemdrive
      elif env_var_ref == "%SYSTEMROOT%":
        value = kb.environ_systemroot
      elif env_var_ref == "%WINDIR%":
        value = kb.environ_windir
      else:
        # We use original match instead of `env_var_ref` as the latter was case
        # corrected.
        expanded += match.group(0)
        continue

      currently_expanded_env_var_refs.add(env_var_ref)
      already_expanded_env_var_refs[env_var_ref] = Expand(value)
      currently_expanded_env_var_refs.remove(env_var_ref)

      expanded += already_expanded_env_var_refs[env_var_ref]

    expanded += unexpanded[offset:]
    return expanded

  kb.environ_path = Expand(kb.environ_path)
  kb.environ_temp = Expand(kb.environ_temp)
  kb.environ_allusersappdata = Expand(kb.environ_allusersappdata)
  kb.environ_allusersprofile = Expand(kb.environ_allusersprofile)
  kb.environ_commonprogramfiles = Expand(kb.environ_commonprogramfiles)
  kb.environ_commonprogramfilesx86 = Expand(kb.environ_commonprogramfilesx86)
  kb.environ_comspec = Expand(kb.environ_comspec)
  kb.environ_driverdata = Expand(kb.environ_driverdata)
  kb.environ_profilesdirectory = Expand(kb.environ_profilesdirectory)
  kb.environ_programfiles = Expand(kb.environ_programfiles)
  kb.environ_programfilesx86 = Expand(kb.environ_programfilesx86)
  kb.environ_programdata = Expand(kb.environ_programdata)
  kb.environ_systemdrive = Expand(kb.environ_systemdrive)
  kb.environ_systemroot = Expand(kb.environ_systemroot)
  kb.environ_windir = Expand(kb.environ_windir)
  return kb


class KnowledgeBaseInterpolation:
  """Interpolation of the given pattern with knowledgebase values.

  Pattern can have placeholder variables like `%%os%%` or `%%fqdn%%` that will
  be replaced by concrete values from the knowledgebase corresponding to these.

  In case of repeated knowledgebase values like `users`, every possible result
  is returned.

  Because interpolation can sometimes omit certain results or use some default
  values, this object exposes a `logs` property with messages when such steps
  were made. These messages can then be forwarded to the user specifying the
  pattern to help the debug issues in case the pattern is behaving unexpectedly.
  """

  def __init__(
      self,
      pattern: str,
      kb: knowledge_base_pb2.KnowledgeBase,
  ) -> None:
    self._results: list[str] = list()
    self._logs: list[str] = list()

    user_attrs = [
        m["attr"] for m in re.finditer(r"%%users\.(?P<attr>\w+)%%", pattern)
    ]
    non_user_attrs = [
        m["attr"] for m in re.finditer(r"%%(?P<attr>\w+)%%", pattern)
    ]

    if not user_attrs:
      # If the pattern does not contain any user attributes, loops below won't
      # yield any results. Hence, we add the pattern as-is for further expansion
      # to always have at least one to work with.
      self._results.append(pattern)
    else:
      # We start with interpolating `users` variables for each user. Because
      # there can be multiple users on the system and the pattern can contain
      # both user and non-user variables we have to then combine all possible
      # user-based interpolations with non-user-based ones.
      for user in kb.users:
        # There might be cases in which username is not strictly necessary but
        # scenario in which we do not have username but have other values is
        # very unlikely. Assuming that users do have usernames makes the logic
        # much simpler below.
        if not (username := user.username):
          self._logs.append(
              f"user {user!r} without username",
          )
          continue

        user_result = pattern

        for attr in user_attrs:
          try:
            value = getattr(user, attr)
          except AttributeError as error:
            raise ValueError(f"`%%users.{attr}%%` does not exist") from error

          if not value:
            if kb.os == "Windows":
              # `userprofile` is a base for all default values so we use various
              # heuristics to derive it in case it is not available.
              if user.userprofile:
                userprofile = user.userprofile
              elif user.homedir:
                userprofile = user.homedir
              elif kb.environ_systemdrive:
                userprofile = f"{kb.environ_systemdrive}\\Users\\{username}"
              else:
                userprofile = f"C:\\Users\\{username}"

              try:
                value = {
                    # pylint: disable=line-too-long
                    # pyformat: disable
                    "userprofile": userprofile,
                    "homedir": userprofile,
                    "temp": f"{userprofile}\\AppData\\Local\\Temp",
                    "desktop": f"{userprofile}\\Desktop",
                    "appdata": f"{userprofile}\\AppData\\Roaming",
                    "localappdata": f"{userprofile}\\AppData\\Local",
                    "cookies": f"{userprofile}\\AppData\\Local\\Microsoft\\Windows\\INetCookies",
                    "recent": f"{userprofile}\\AppData\\Roaming\\Microsoft\\Windows\\Recent",
                    "personal": f"{userprofile}\\Documents",
                    "startup": f"{userprofile}\\AppData\\Roaming\\Microsoft\\Windows\\Start Menu\\Programs\\Startup",
                    # pylint: enable=line-too-long
                    # pyformat: enable
                }[attr]
              except KeyError:
                self._logs.append(
                    f"user {username!r} is missing {attr!r} (no Windows "
                    "default available)",
                )
                break
            else:
              self._logs.append(
                  f"user {username!r} is missing {attr!r} (no fallback "
                  "available)",
              )
              break

            self._logs.append(
                f"using default {value!r} for {attr!r} for user {username!r}",
            )

          user_result = user_result.replace(f"%%users.{attr}%%", value)
        else:
          # This will run only if we successfully filled every variable. If any
          # is missing we will break the loop and this block won't be executed.
          self._results.append(user_result)

    # At this point all results have no user variables, so there is only one way
    # to interpolate them. We do a pass for every variable in every result to
    # expand these.
    for attr in non_user_attrs:
      try:
        value = getattr(kb, attr)
      except AttributeError as error:
        raise ValueError(f"`%%{attr}%%` does not exist") from error

      if not value:
        self._logs.append(
            f"{attr!r} is missing",
        )
        # If the attribute value is missing in the knowledge base, the pattern
        # cannot be interpolated and should yield no results.
        self._results = []

      # Because strings in Python are immutable, we cannot simply iterate over
      # the elements of the list if we want to update them, so we use indices to
      # simulate references.
      for i in range(len(self._results)):
        self._results[i] = self._results[i].replace(f"%%{attr}%%", value)

  @property
  def results(self) -> Sequence[str]:
    return self._results

  @property
  def logs(self) -> Sequence[str]:
    return self._logs



#!/usr/bin/env python
"""Configuration parameters for the artifact subsystem."""


from grr_response_core.lib import config_lib

config_lib.DEFINE_list("Artifacts.artifact_dirs", [
    "%(grr_response_core/artifacts@grr-response-core|resource)",
    "%(grr_response_core/artifacts/local@grr-response-core|resource)"
], "A list directories to load artifacts from.")

config_lib.DEFINE_list(
    "Artifacts.netgroup_filter_regexes", [],
    help="Only parse groups that match one of these regexes"
    " from /etc/netgroup files.")

config_lib.DEFINE_list(
    "Artifacts.netgroup_ignore_users", [],
    help="Exclude these users when parsing /etc/netgroup "
    "files.")
